package fem;
import java.applet.Applet;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.io.InputStream;
import java.net.URL;
import fem.component.FEMActionManager;
import fem.component.FEMViewCanvas;
import fem.io.FEMModelReader;
import figure.matrix.MatrixSolveMethod;

//--------------------------------------------------------------------//
/**
 * FEM Avbg CNX
 * @author aG
 * @version 1.1
 */
public class FEMApplet extends Applet{

/** G[bZ[W */
  protected String message;

//--------------------------------------------------------------------//
/**
 * Avbg
 */
  public void init(){
    setLayout(new BorderLayout());
    Color bgColor=getTagColor("backgroundcolor");
    Color fgColor=getTagColor("foregroundcolor");
    FEMDataModel model=getDataModel();
    model.setAnalyzeType(getAnalyzeType());
    model.setMethod(getMatrixSolveMethod());
    FEMViewCanvas canvas=new FEMViewCanvas
      (model,message,getTagFont(),bgColor,fgColor);
    FEMActionManager manager=new FEMActionManager(canvas);
    add(canvas,BorderLayout.CENTER);
    add(manager.getControlPanel(),BorderLayout.SOUTH);
  }

//--------------------------------------------------------------------//
/**
 * \[Xj
 */
  public void destroy(){
    removeAll();
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽ FEM f[^Ԃ
 * @return HTML^OŎw肳ꂽ FEM f[^
 */
  public FEMDataModel getDataModel(){
    String dataName=getParameter("data");
    if(dataName==null) dataName="sample.dat";
    message=null;
    try{
      Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
    }
    catch(SecurityException ex){
      ex.printStackTrace();
      message=String.valueOf(ex);
      return null;
    }
    FEMDataModel model=new FEMDataModel();
    try{
      FEMModelReader reader=new FEMModelReader(model);
      reader.read(new URL(getDocumentBase(),dataName));
      return model;
    }
    catch(Exception ex){
      ex.printStackTrace();
      message=String.valueOf(ex);
      return null;
    }
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽFԂ
 * @param name p[^
 * @return HTML^OŎw肳ꂽF
 */
  protected Color getTagColor(String name){
    String value=getParameter(name);
    if(value==null) return null;
    try{
      return new Color(Integer.parseInt(value,16));
    }
    catch(NumberFormatException ex){
      return null;
    }
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽtHgԂ
 * @return HTML^OŎw肳ꂽtHg
 */
  protected Font getTagFont(){
    String fontName=getParameter("fontname");
    if(fontName==null) fontName="Serif";

    String s=getParameter("fontstyle");
    int fontStyle=Font.PLAIN;
    if(s!=null){
      if(s.equals("bold")){
      	fontStyle=Font.BOLD;
      }
      else if(s.equals("italic")){
      	fontStyle=Font.ITALIC;
      }
    }

    s=getParameter("fontsize");
    int fontSize=14;
    try{
      fontSize=Integer.parseInt(s);
    }
    catch(NumberFormatException ex){
    }

    return new Font(fontName,fontStyle,fontSize);
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽ͎ނԂ
 * @return HTML^OŎw肳ꂽ͎
 */
  protected AnalyzeType getAnalyzeType(){
    AnalyzeType type=AnalyzeType.STRUCTURE;
    String value=getParameter("type");
    if(value==null){
      return type;
    }
    else if(value.equals(AnalyzeType.HEAT_TRANSFER.getName())){
      type=AnalyzeType.HEAT_TRANSFER;
    }
    return type;
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽs@Ԃ
 * @return HTML^OŎw肳ꂽs@
 */
  protected MatrixSolveMethod getMatrixSolveMethod(){
    MatrixSolveMethod method=MatrixSolveMethod.ILUCG;
    String value=getParameter("method");
    if(value==null){
      return method;
    }
    else if(value.equals(MatrixSolveMethod.GAUSS.getName())){
      method=MatrixSolveMethod.GAUSS;
    }
    else if(value.equals(MatrixSolveMethod.GAUSS_JORDAN.getName())){
      method=MatrixSolveMethod.GAUSS_JORDAN;
    }
    else if(value.equals(MatrixSolveMethod.LU.getName())){
      method=MatrixSolveMethod.LU;
    }
    else if(value.equals(MatrixSolveMethod.SOR.getName())){
      method=MatrixSolveMethod.SOR;
    }
    else if(value.equals(MatrixSolveMethod.CG.getName())){
      method=MatrixSolveMethod.CG;
    }
    return method;
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽlԂ
 * @param name p[^
 * @return HTML^OŎw肳ꂽl
 */
  protected double getTagNumber(String name){
    String value=getParameter(name);
    if(value==null) return Double.NaN;
    try{
      return Double.parseDouble(value);
    }
    catch(NumberFormatException ex){
      return Double.NaN;
    }
  }

//--------------------------------------------------------------------//
/**
 * AvbgԂ
 * @return Avbg܂ޕ
 */
  public String getAppletInfo(){
    return "Applet of 3D FEM\nAuthor K.Higuchi";
  }

//--------------------------------------------------------------------//
/**
 * ̃Avbg߂p[^Ԃ
 * @return Avbg߂p[^i[z
 */
  public String[][] getParameterInfo(){
    return new String[][]
      {{"data","String","FEM data file name."},
       {"fontname","String","Font name. Default is Serif."},
       {"fontstyle","int","Font style. Default is Font.PLAIN."},
       {"fontsize","int","Font size. Default is 14."},
       {"backgroundcolor","hexadecimal RGB number",
      	"The color of background."},
       {"foregroundcolor","hexadecimal RGB number",
      	"The color of foreground."},
       {"type","String","The type of analyze."},
       {"method","String",
      	"The method for solving systems of linear equations."}};
  }
}
