package fem;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import fem.bc.BoundaryCondition;
import fem.bc.HeatTransferBound;
import fem.bc.Load;
import fem.bc.Pressure;
import fem.bc.Restraint;
import fem.bc.Temperature;
import fem.border.AbstractElementBorder;
import fem.element.TriElement1;
import fem.result.Result;
import fem.result.Strain;
import fem.result.Stress;
import figure.Matrix;
import figure.Point3D;
import figure.Vector;
import figure.Vector3D;
import figure.Vector3DR;
import figure.matrix.SparseMatrixSolver;
import figure.matrix.MatrixSolveException;
import figure.matrix.MatrixSolveMethod;
import figure.matrix.MatrixSolver;

//--------------------------------------------------------------------//
/**
 * FEM f[^f
 */
public class FEMDataModel{

/** bVf */
  protected MeshModel mesh;

/** ޗi[郊Xg */
  protected ArrayList<Material> materials;

/** VFp[^i[郊Xg */
  protected ArrayList<ShellParameter> shellParams;

/** E */
  protected BoundaryCondition bc;

/** vZ */
  protected Result result;

/** f̗L\ boolean ^ */
  protected boolean modelExist;

/** }gbNX */
  protected MatrixSolver matrix;

/** ׏dxNg */
  protected Vector force;

/** f̎Rx */
  protected int dof;

/** ͎ */
  protected AnalyzeType type;

/** s@ */
  protected MatrixSolveMethod method;

/**
 * FEM f[^f\z
 */
  public FEMDataModel(){
    materials=new ArrayList<Material>();
    shellParams=new ArrayList<ShellParameter>();
    mesh=new MeshModel();
    bc=new BoundaryCondition();
    result=new Result();
    modelExist=false;
    dof=0;
    type=AnalyzeType.STRUCTURE;
    method=MatrixSolveMethod.ILUCG;
  }

//--------------------------------------------------------------------//
/**
 * bVfԂ
 * @return bVf
 */
  public MeshModel getMesh(){
    return mesh;
  }

//--------------------------------------------------------------------//
/**
 * EԂ
 * @return E
 */
  public BoundaryCondition getBoundaryCondition(){
    return bc;
  }

//--------------------------------------------------------------------//
/**
 * vZʂԂ
 * @return vZ
 */
  public Result getResult(){
    return result;
  }

//--------------------------------------------------------------------//
/**
 * ޗ̐Ԃ
 * @return ޗ̐
 */
  public int materialCount(){
    return materials.size();
  }

//--------------------------------------------------------------------//
/**
 * ޗԂ
 * @param index ޗ̃CfbNX
 * @return w肳ꂽʒu̍ޗ
 */
  public Material getMaterial(int index){
    return materials.get(index);
  }

//--------------------------------------------------------------------//
/**
 * ޗǉ
 * @param material ޗ
 */
  public void addMaterial(Material material){
    materials.add(material);
  }

//--------------------------------------------------------------------//
/**
 * VFp[^̐Ԃ
 * @return VFp[^̐
 */
  public int shellParameterCount(){
    return shellParams.size();
  }

//--------------------------------------------------------------------//
/**
 * VFp[^Ԃ
 * @param index VFp[^̃CfbNX
 * @return w肳ꂽʒũVFp[^
 */
  public ShellParameter getShellParameter(int index){
    return shellParams.get(index);
  }

//--------------------------------------------------------------------//
/**
 * VFp[^ǉ
 * @param param VFp[^
 */
  public void addShellParameter(ShellParameter param){
    shellParams.add(param);
  }

//--------------------------------------------------------------------//
/**
 * ͎ނԂ
 * @return ͎
 */
  public AnalyzeType getAnalyzeType(){
    return type;
  }

//--------------------------------------------------------------------//
/**
 * ͎ނݒ肷
 * @param type ͎
 */
  public void setAnalyzeType(AnalyzeType type){
    this.type=type;
  }

//--------------------------------------------------------------------//
/**
 * s@ݒ肷
 * @param method s@
 */
  public void setMethod(MatrixSolveMethod method){
    this.method=method;
  }

//--------------------------------------------------------------------//
/**
 * f[^
 */
  public void clear(){
    materials.clear();
    shellParams.clear();
    modelExist=false;
    mesh.clear();
    bc.clear();
    dof=0;
    clearResults();
  }

//--------------------------------------------------------------------//
/**
 * vZʂ
 */
  public void clearResults(){
    result.clear();
  }

//--------------------------------------------------------------------//
/**
 * f̗L\ boolean ^Ԃ
 * @return f݂ꍇ true
 */
  public boolean modelExist(){
    return modelExist;
  }

//--------------------------------------------------------------------//
/**
 * f
 */
  public void init() throws LabelException{
    materials.trimToSize();
    Collections.sort(materials,MeshModel.LABEL_COMPARATOR);
    mesh.init();
    bc.init();

    reNumbering();
    resetMaterialLabel();
    resetShellParameterLabel();
    mesh.getFreeFaces();
    setNodeDoF();
    modelExist=true;
  }

//--------------------------------------------------------------------//
/**
 * ߓ_|C^ݒ肷
 * @throws LabelException f[^xvȂƂ
 */
  public void reNumbering() throws LabelException{
    LabelMap labelMap=new LabelMap();
    int nodeCount=mesh.getNodeCount();
    for(int i=0;i<nodeCount;i++){
      labelMap.put(mesh.getNode(i),i);
    }

    int elementCount=mesh.getElementCount();
    for(int i=0;i<elementCount;i++){
      labelMap.reset((Nodes)mesh.getElement(i));
    }

    int restCount=bc.restraintCount();
    for(int i=0;i<restCount;i++){
      labelMap.reset(bc.getRestraint(i));
    }

    int loadCount=bc.loadCount();
    for(int i=0;i<loadCount;i++){
      labelMap.reset(bc.getLoad(i));
    }

    int tempCount=bc.temperatureCount();
    for(int i=0;i<tempCount;i++){
      labelMap.reset(bc.getTemperature(i));
    }

    labelMap.clear();
    for(int i=0;i<elementCount;i++){
      labelMap.put(mesh.getElement(i),i);
    }

    int pressureCount=bc.pressureCount();
    for(int i=0;i<pressureCount;i++){
      labelMap.reset(bc.getPressure(i));
    }

    int htcCount=bc.htcCount();
    for(int i=0;i<htcCount;i++){
      labelMap.reset(bc.getHTBound(i));
    }
  }

//--------------------------------------------------------------------//
/**
 * ޗ|C^ݒ肷
 * @throws LabelException f[^xvȂƂ
 */
  public void resetMaterialLabel() throws LabelException{
    int materialCount=materials.size();
    if(materialCount==0){
      materials.add(new Material(1));
      return;
    }

    HashMap<Integer,Integer> materialMap=new HashMap<Integer,Integer>();
    for(int i=0;i<materialCount;i++){
      materialMap.put(getMaterial(i).label(),i);
    }

    int elementCount=mesh.getElementCount();
    for(int i=0;i<elementCount;i++){
      Element element=mesh.getElement(i);
      Integer newMaterial=materialMap.get(element.getMaterial());
      if(newMaterial!=null){
      	element.setMaterial(newMaterial.intValue());
      }
      else{
      	throw new LabelException("Bad material label");
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * VFp[^|C^ݒ肷
 * @throws LabelException f[^xvȂƂ
 */
  public void resetShellParameterLabel() throws LabelException{
    int shellParamCount=shellParams.size();
    if(shellParamCount==0) return;

    HashMap<Integer,Integer> paramMap=new HashMap<Integer,Integer>();
    for(int i=0;i<shellParamCount;i++){
      paramMap.put(getShellParameter(i).label(),i);
    }

    int elementCount=mesh.getElementCount();
    for(int i=0;i<elementCount;i++){
      Element element=mesh.getElement(i);
      if(element instanceof ShellElement){
      	ShellElement se=(ShellElement)element;
      	Integer newParam=paramMap.get(se.getParam());
      	if(newParam!=null){
      	  se.setParam(newParam.intValue());
      	}
      	else{
      	  throw new LabelException("Bad material label");
      	}
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * ߓ_̎Rxݒ肷
 */
  protected void setNodeDoF(){
    int nodeCount=mesh.getNodeCount();
    int elementCount=mesh.getElementCount();
    int dof[]=new int[nodeCount];
    Arrays.fill(dof,3);
    for(int i=0;i<elementCount;i++){
      Element element=mesh.getElement(i);
      if(element instanceof ShellElement){
      	int nn=element.nodeCount();
      	for(int j=0;j<nn;j++){
      	  dof[element.node(j)]=6;
      	}
      }
    }
    this.dof=bc.bcList(mesh.getNodeCount(),dof);
  }

//--------------------------------------------------------------------//
/**
 * vZ
 */
  public void calculate(){
    switch(type){
      case STRUCTURE:
      	createStiffnessMatrix();
      	break;
      case HEAT_TRANSFER:
      	createHeatMatrix();
      	break;
    }
    try{
      Vector vector=matrix.solve(force,method);
      result.clear();
      switch(type){
      	case STRUCTURE:
      	  result.setDisplacement(bc,vector,bc.getRestraints());
      	  calculateStress();
      	  break;
      	case HEAT_TRANSFER:
      	  result.setTemperature(bc,vector,bc.getTemperature());
      	  break;
      }
    }
    catch(MatrixSolveException ex){
      ex.printStackTrace();
    }
  }

//--------------------------------------------------------------------//
/**
 * }gbNXE׏dxNg쐬
 */
  public void createStiffnessMatrix(){
    int length=0;
    for(int i=0;i<bc.bcList.length;i++){
      if(bc.bcList[i]<0) length++;
    }
    int ii=0;
    int reducedList[]=new int[length];
    for(int i=0;i<bc.bcList.length;i++){
      if(bc.bcList[i]<0){
      	reducedList[ii]=i;
      	ii++;
      }
    }

// }gbNX̍쐬
    MatrixSolver matrix1=stiffnessMatrix();
    matrix=matrix1.extruct(length,reducedList,reducedList);

// ׏dxNg̍쐬
    Vector force1=bc.loadVector(dof);
    int pressureCount=bc.pressureCount();
    for(int i=0;i<pressureCount;i++){
      Pressure pressure=bc.getPressure(i);
      Element element=mesh.getElement(pressure.label());
      Point3D center=element.center(mesh.nodes(element));
      ElementBorder border=pressure.getBorder(element);
      int nodeCount=border.nodeCount();
      Point3D p[]=mesh.nodes(border);
      Vector ps=border.shapeFunctionVector(p,pressure.press);
      Vector3D norm=AbstractElementBorder.normalVector(p);
      if(norm.dotProduct(new Vector3D(p[0],center))<0.0){
      	norm.mul(-1.0);
      }
      else if(element instanceof ShellElement){
      	norm.mul(-1.0);
      }
      for(int j=0;j<nodeCount;j++){
      	int index0=bc.nodeIndex(border.node(j));
      	double press=ps.get(j);
      	force1.add(index0,press*norm.x);
      	force1.add(index0+1,press*norm.y);
      	force1.add(index0+2,press*norm.z);
      }
    }
    for(int i=0;i<bc.bcList.length;i++){
      if(bc.bcList[i]>=0){
      	Restraint rest=bc.getRestraint(bc.bcList[i]/6);
      	switch(bc.bcList[i]%6){
      	  case 0:
      	    for(int j=0;j<force1.size;j++){
      	      force1.add(j,-rest.x*matrix1.get(j,i));
      	    }
      	    break;
      	  case 1:
      	    for(int j=0;j<force1.size;j++){
      	      force1.add(j,-rest.y*matrix1.get(j,i));
      	    }
      	    break;
      	  case 2:
      	    for(int j=0;j<force1.size;j++){
      	      force1.add(j,-rest.z*matrix1.get(j,i));
      	    }
      	    break;
      	  case 3:
      	    for(int j=0;j<force1.size;j++){
      	      force1.add(j,-rest.rx*matrix1.get(j,i));
      	    }
      	    break;
      	  case 4:
      	    for(int j=0;j<force1.size;j++){
      	      force1.add(j,-rest.ry*matrix1.get(j,i));
      	    }
      	    break;
      	  case 5:
      	    for(int j=0;j<force1.size;j++){
      	      force1.add(j,-rest.rz*matrix1.get(j,i));
      	    }
      	    break;
      	}
      }
    }
    force=new Vector(length);
    for(int i=0;i<length;i++){
      force.set(i,force1.get(reducedList[i]));
    }
  }

//--------------------------------------------------------------------//
/**
 * }gbNX쐬
 * @return }gbNX
 */
  public MatrixSolver stiffnessMatrix(){
    int elementCount=mesh.getElementCount();
    MatrixSolver matrix=new SparseMatrixSolver(dof);
    for(int k=0;k<elementCount;k++){
      Element element=mesh.getElement(k);
      int nodeCount=element.nodeCount();
      int nodesOnElement[]=element.nodes();
      if(element instanceof ShellElement){
      	ShellElement se=(ShellElement)element;
      	ShellParameter sp=getShellParameter(se.getParam());
      	if(element instanceof TriElement1){
      	  getMaterial(se.getMaterial()).matrix2Dstress();
      	}
      	else{
      	  getMaterial(se.getMaterial()).matrixShell();
      	}
      	Matrix km=se.stiffnessMatrix
      	  (mesh.nodes(se),getMaterial(se.getMaterial()),sp.thickness());
      	for(int i=0;i<nodeCount;i++){
      	  for(int j=0;j<nodeCount;j++){
      	    int row0=bc.nodeIndex(nodesOnElement[i]);
      	    int column0=bc.nodeIndex(nodesOnElement[j]);
      	    int i0=6*i;
      	    int j0=6*j;
      	    for(int i1=0;i1<6;i1++){
      	      for(int j1=0;j1<6;j1++){
      	      	matrix.add(row0+i1,column0+j1,km.get(i0+i1,j0+j1));
      	      }
      	    }
      	  }
      	}
      }
      else{
      	getMaterial(element.getMaterial()).matrix3D();
      	Matrix km=element.stiffnessMatrix
      	  (mesh.nodes(element),getMaterial(element.getMaterial()));
      	for(int i=0;i<nodeCount;i++){
      	  for(int j=0;j<nodeCount;j++){
      	    int row0=bc.nodeIndex(nodesOnElement[i]);
      	    int column0=bc.nodeIndex(nodesOnElement[j]);
      	    int i0=3*i;
      	    int j0=3*j;
      	    for(int i1=0;i1<3;i1++){
      	      for(int j1=0;j1<3;j1++){
      	      	matrix.add(row0+i1,column0+j1,km.get(i0+i1,j0+j1));
      	      }
      	    }
      	  }
      	}
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * MvZ̃}gbNXExNgvZ
 */
  protected void createHeatMatrix(){
    int nodeCount=mesh.getNodeCount();
    int elementCount=mesh.getElementCount();

    int dof=bc.bcList(mesh.getNodeCount());
    int length=nodeCount-dof;
    int ii=0;
    int reducedList[]=new int[length];
    for(int i=0;i<bc.bcList.length;i++){
      if(bc.bcList[i]<0){
      	reducedList[ii]=i;
      	ii++;
      }
    }

    SparseMatrixSolver matrix1=new SparseMatrixSolver(nodeCount);
    Vector vector1=bc.tempVector(nodeCount);
    for(int i=0;i<elementCount;i++){
      if(mesh.getElement(i) instanceof SolidElement){
      	SolidElement element=(SolidElement)mesh.getElement(i);
      	nodeCount=element.nodeCount();
      	int nodesOnElement[]=element.nodes();
      	Material material=materials.get(element.getMaterial());
      	Matrix ls=element.gradMatrix(mesh.nodes(element),
      	      	      	      	     material.heatConductivity);
      	for(int i1=0;i1<nodeCount;i1++){
      	  for(int i2=0;i2<nodeCount;i2++){
      	    matrix1.add(nodesOnElement[i1],nodesOnElement[i2],
      	      	      	ls.get(i1,i2));
      	  }
      	}
      }
    }

    int boundCount=bc.htcCount();
    for(int i=0;i<boundCount;i++){
      HeatTransferBound htc=bc.getHTBound(i);
      ElementBorder border=htc.getBorder(mesh.getElement(htc.label()));
      nodeCount=border.nodeCount();
      int nodesOnElement[]=border.nodes();
      Node nodes[]=mesh.nodes(border);
      Matrix hm=border.shapeFunctionMatrix(nodes,htc.htc);
      Vector hv=border.shapeFunctionVector(nodes,htc.outTemp*htc.htc);
      for(int i1=0;i1<nodeCount;i1++){
      	for(int i2=0;i2<nodeCount;i2++){
      	  matrix1.add(nodesOnElement[i1],nodesOnElement[i2],
      	      	      hm.get(i1,i2));
      	}
      	vector1.add(nodesOnElement[i1],hv.get(i1));
      }
    }

    for(int i=0;i<bc.bcList.length;i++){
      if(bc.bcList[i]>=0){
      	Temperature temp=bc.getTemperature(bc.bcList[i]);
      	for(int j=0;j<vector1.size;j++){
      	  vector1.add(j,-temp.temperature()*matrix1.get(j,i));
      	}
      }
    }

    matrix=matrix1.extruct(length,reducedList,reducedList);
    force=new Vector(length);
    for(int i=0;i<length;i++){
      force.set(i,vector1.get(reducedList[i]));
    }
  }

//--------------------------------------------------------------------//
/**
 * ߓ_͂vZ
 */
  public void calculateStress(){
    int nodeCount=mesh.getNodeCount();
    int elementCount=mesh.getElementCount();
    Strain vstrain1[]=new Strain[nodeCount];
    Strain vstrain2[]=new Strain[nodeCount];
    Stress vstress1[]=new Stress[nodeCount];
    Stress vstress2[]=new Stress[nodeCount];
    double angle[]=new double[nodeCount];
    for(int i=0;i<nodeCount;i++){
      vstrain1[i]=new Strain();
      vstrain2[i]=new Strain();
      vstress1[i]=new Stress();
      vstress2[i]=new Stress();
      angle[i]=0.0;
    }
    for(int i=0;i<elementCount;i++){
      Element element=mesh.getElement(i);
      int nodesOnElement[]=element.nodes();
      Point3D p[]=new Point3D[nodesOnElement.length];
      Vector3DR v[]=new Vector3DR[nodesOnElement.length];
      for(int j=0;j<nodesOnElement.length;j++){
      	p[j]=mesh.getNode(nodesOnElement[j]);
      	v[j]=result.getDisplacement(nodesOnElement[j]);
      }
      Material material=getMaterial(element.getMaterial());
      double eangle[]=element.angle(p);
      if(element instanceof ShellElement){
      	if(element instanceof TriElement1){
      	  material.matrix2Dstress();
      	}
      	else{
      	  material.matrixShell();
      	}
      	ShellElement shell=(ShellElement)element;
      	ShellParameter sp=getShellParameter(shell.getParam());
      	Strain eps1[]=shell.strain(p,v,1.0,sp.thickness());
      	Stress str1[]=shell.stress(p,v,material.matrix(),1.0,
      	      	      	      	   sp.thickness());
      	Strain eps2[]=shell.strain(p,v,-1.0,sp.thickness());
      	Stress str2[]=shell.stress(p,v,material.matrix(),-1.0,
      	      	      	      	   sp.thickness());
      	for(int j=0;j<nodesOnElement.length;j++){
      	  eps1[j].mul(eangle[j]);
      	  str1[j].mul(eangle[j]);
      	  eps2[j].mul(eangle[j]);
      	  str2[j].mul(eangle[j]);
      	  vstrain1[nodesOnElement[j]].add(eps1[j]);
      	  vstress1[nodesOnElement[j]].add(str1[j]);
      	  vstrain2[nodesOnElement[j]].add(eps2[j]);
      	  vstress2[nodesOnElement[j]].add(str2[j]);
      	  angle[nodesOnElement[j]]+=eangle[j];
      	}
      }
      else if(element instanceof SolidElement){
      	material.matrix3D();
      	SolidElement solid=(SolidElement)element;
      	Strain eps[]=solid.strain(p,v);
      	Stress str[]=solid.stress(eps,material.matrix());
      	for(int j=0;j<nodesOnElement.length;j++){
      	  eps[j].mul(eangle[j]);
      	  str[j].mul(eangle[j]);
      	  vstrain1[nodesOnElement[j]].add(eps[j]);
      	  vstress1[nodesOnElement[j]].add(str[j]);
      	  vstrain2[nodesOnElement[j]].add(eps[j]);
      	  vstress2[nodesOnElement[j]].add(str[j]);
      	  angle[nodesOnElement[j]]+=eangle[j];
      	}
      }
    }
    for(int i=0;i<nodeCount;i++){
      if(angle[i]==0.0){
      	result.addStrain(new Strain(),new Strain());
      	result.addStress(new Stress(),new Stress());
      }
      else{
      	double ainv=1.0/angle[i];
      	vstrain1[i].mul(ainv);
      	vstress1[i].mul(ainv);
      	vstrain2[i].mul(ainv);
      	vstress2[i].mul(ainv);
      	result.addStrain(vstrain1[i],vstrain2[i]);
      	result.addStress(vstress1[i],vstress2[i]);
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @return }gbNX
 */
  public MatrixSolver getStiffnessMatrix(){
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * ׏dxNgԂ
 * @return ׏dxNg
 */
  public Vector getForce(){
    return force;
  }
}
