package fem.border;
import java.util.Arrays;
import fem.AbstractNodes;
import fem.ElementBorder;
import fem.Node;
import figure.DefaultMatrix;
import figure.Matrix;
import figure.Point3D;
import figure.Vector;
import figure.Vector3D;

//--------------------------------------------------------------------//
/**
 * vfE̒ۃX[p[NX
 */
public abstract class AbstractElementBorder extends AbstractNodes
  implements Cloneable,ElementBorder{

/** vfx */
  public int element;

  protected AbstractElementBorder(int element){
    this.element=element;
  }

//--------------------------------------------------------------------//
/**
 * vfÊԂ
 * @return vfE
 */
  public abstract String getName();

//--------------------------------------------------------------------//
/**
 * 񏇂ɕ񂾐ߓ_xԂ
 * @return ߓ_x񏇂ɕׂĊi[z
 */
  public int[] cycleNodes(){
    return nodes;
  }

//--------------------------------------------------------------------//
/**
 * vfxԂ
 * @return vfx
 */
  public int element(){
    return element;
  }

//--------------------------------------------------------------------//
/**
 * vfxĐݒ肷
 * @param element vfx
 */
  public void resetElementLabel(int element){
    this.element=element;
  }

//--------------------------------------------------------------------//
/**
 * @xNgԂ
 * @param p _i[z
 * @return @xNg
 */
  public static Vector3D normalVector(Point3D p[]){
    if(p.length<3){
      return null;
    }
    else if((p.length==3) || (p.length==6)){
      Vector3D norm=Vector3D.crossProduct
      	(new Vector3D(p[0],p[1]),new Vector3D(p[0],p[2]));
      norm.normalize();
      return norm;
    }
    else if((p.length==4) || (p.length==8)){
      Vector3D norm=Vector3D.crossProduct
      	(new Vector3D(p[0],p[2]),new Vector3D(p[1],p[3]));
      norm.normalize();
      return norm;
    }
    else{
      double vx=0.0;
      double vy=0.0;
      double vz=0.0;
      for(int i=0;i<p.length;i++){
      	Point3D p1=p[(i+1)%p.length];
      	Point3D p2=p[(i+2)%p.length];
      	Vector3D norm=Vector3D.crossProduct
      	  (new Vector3D(p[i],p1),new Vector3D(p[i],p2));
      	vx+=norm.x;
      	vy+=norm.y;
      	vz+=norm.z;
      }
      Vector3D norm=new Vector3D(vx,vy,vz);
      norm.normalize();
      return norm;
    }
  }

//--------------------------------------------------------------------//
/**
 * [RrsԂ
 * @param p ߓ_i[z
 * @param sf `֐s
 * @return [Rrs
 */
  protected Matrix jacobianMatrix(Point3D p[],Matrix sf){
    Matrix jacob=new DefaultMatrix(2,3);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      double dndx2=sf.get(i,1);
      double dndy2=sf.get(i,2);
      jacob.add(0,0,dndx2*p[i].x);
      jacob.add(0,1,dndx2*p[i].y);
      jacob.add(0,2,dndx2*p[i].z);
      jacob.add(1,0,dndy2*p[i].x);
      jacob.add(1,1,dndy2*p[i].y);
      jacob.add(1,2,dndy2*p[i].z);
    }
    return jacob;
  }

//--------------------------------------------------------------------//
/**
 * [RrAԂ
 * @param ja Rrs
 * @return RrA
 */
  protected static double jacobian(Matrix ja){
    double j1=ja.get(0,1)*ja.get(1,2)-ja.get(0,2)*ja.get(1,1);
    double j2=ja.get(0,2)*ja.get(1,0)-ja.get(0,0)*ja.get(1,2);
    double j3=ja.get(0,0)*ja.get(1,1)-ja.get(0,1)*ja.get(1,0);
    return Math.sqrt(j1*j1+j2*j2+j3*j3);
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX { <i>N<sub>i</sub>N<sub>j</sub></i> } Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @return `֐}gbNX
 */
  protected Matrix shapeFunctionMatrix(Point3D p[],double xsi,
      	      	      	      	       double eta){
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,sf);
    double jacobian=jacobian(ja);
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(nodeCount);
    for(int i=0;i<nodeCount;i++){
      for(int j=0;j<nodeCount;j++){
      	matrix.set(i,j,jacobian*sf.get(i,0)*sf.get(j,0));
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * `֐xNg { <i>N<sub>i</sub></i> } Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @return `֐xNg
 */
  protected Vector shapeFunctionVector(Point3D p[],double xsi,
      	      	      	      	       double eta){
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,sf);
    double jacobian=jacobian(ja);
    int nodeCount=nodeCount();
    Vector vector=new Vector(nodeCount);
    for(int i=0;i<nodeCount;i++){
      vector.set(i,jacobian*sf.get(i,0));
    }
    return vector;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃Rs[Ԃ
 * @return ̃CX^X̕
 */
  public Object clone(){
    return copyBorder();
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃Rs[Ԃ
 * @return ̃CX^X̕
 */
  public ElementBorder copyBorder(){
    try{
      AbstractElementBorder border=(AbstractElementBorder)super.clone();
      border.nodes=new int[nodes.length];
      System.arraycopy(nodes,0,border.nodes,0,nodes.length);
      return border;
    }
    catch(CloneNotSupportedException ex){
      throw new InternalError();
    }
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃nbVR[hlԂ
 * @return IuWFNg̃nbVR[hl
 */
  public int hashCode(){
    int p[]=(int[])nodes().clone();
    Arrays.sort(p);
    int bit=0;
    for(int i=0;i<p.length;i++){
      bit=31*bit+p[i];
    }
    return bit;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @param obj rΏۂ̃IuWFNg
 * @return rΏۂ̃IuWFNg <code>AbstractElementBorder</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof AbstractElementBorder){
      AbstractElementBorder eb=(AbstractElementBorder)obj;
      int p1[]=(int[])nodes().clone();
      int p2[]=(int[])eb.nodes().clone();
      if(p1.length!=p2.length) return false;
      Arrays.sort(p1);
      Arrays.sort(p2);
      return Arrays.equals(p1,p2);
    }
    else{
      return false;
    }
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @return ̃IuWFNg̕\
 */
  public String toString(){
    StringBuilder sb=new StringBuilder(getName());
    sb.append('\t');
    sb.append(element);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      sb.append('\t');
      sb.append(nodes[i]);
    }
    return sb.toString();
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @param p ߓ_i[z
 * @return ̃IuWFNg̕\
 */
  public String toString(Node p[]){
    StringBuilder sb=new StringBuilder(getName());
    sb.append('\t');
    sb.append(element);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      sb.append('\t');
      sb.append(p[i].label());
    }
    return sb.toString();
  }
}
