package fem.border;
import fem.ElementBorder;
import figure.DefaultMatrix;
import figure.Matrix;
import figure.Point3D;
import figure.Vector;

//--------------------------------------------------------------------//
/**
 * lp`2vfENX
 */
public class QuadangleBorder2 extends AbstractElementBorder{

/** vfE */
  public static final String BORDER_NAME="QuadangleBorder2";

/** ߓ_ */
  public static final int NODE_COUNT=8;

  public QuadangleBorder2(int element,int p1,int p2,int p3,int p4,
      	      	      	  int p5,int p6,int p7,int p8){
    super(element);
    nodes=new int[NODE_COUNT];
    nodes[0]=p1;
    nodes[1]=p2;
    nodes[2]=p3;
    nodes[3]=p4;
    nodes[4]=p5;
    nodes[5]=p6;
    nodes[6]=p7;
    nodes[7]=p8;
  }

  public QuadangleBorder2(int element,int p[]){
    this(element,p[0],p[1],p[2],p[3],p[4],p[5],p[6],p[7]);
  }

//--------------------------------------------------------------------//
/**
 * vfÊԂ
 * @return vfE
 */
  public String getName(){
    return BORDER_NAME;
  }

//--------------------------------------------------------------------//
/**
 * ߓ_Ԃ
 * @return ߓ_
 */
  public int nodeCount(){
    return NODE_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * 񏇂ɕ񂾐ߓ_xԂ
 * @return ߓ_x񏇂ɕׂĊi[z
 */
  public int[] cycleNodes(){
    return new int[]{nodes[0],nodes[4],nodes[1],nodes[5],nodes[2],
      	      	     nodes[6],nodes[3],nodes[7]};
  }

//--------------------------------------------------------------------//
/**
 * vfE𕪊
 * @return vfEi[z
 */
  public ElementBorder[] splitBorder(){
    return new ElementBorder[]
      {new TriangleBorder1(element,nodes[0],nodes[4],nodes[7]),
       new TriangleBorder1(element,nodes[1],nodes[5],nodes[4]),
       new TriangleBorder1(element,nodes[2],nodes[6],nodes[5]),
       new TriangleBorder1(element,nodes[3],nodes[7],nodes[6]),
       new QuadangleBorder1(element,nodes[4],nodes[5],nodes[6],
      	      	      	    nodes[7])};
  }

//--------------------------------------------------------------------//
/**
 * `֐s<br>
 * { <i>N<sub>i</sub></i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i> }<br>
 * Ԃ
 * @param xsi <i></i> W
 * @param eta <i></i> W
 */
  public Matrix shapeFunction(double xsi,double eta){
    Matrix m=new DefaultMatrix(NODE_COUNT,3);
    m.set(0,0,0.25*(1.0-xsi)*(1.0-eta)*(-xsi-eta-1.0));
    m.set(1,0,0.25*(1.0+xsi)*(1.0-eta)*(xsi-eta-1.0));
    m.set(2,0,0.25*(1.0+xsi)*(1.0+eta)*(xsi+eta-1.0));
    m.set(3,0,0.25*(1.0-xsi)*(1.0+eta)*(-xsi+eta-1.0));
    m.set(4,0,0.5*(1.0+xsi)*(1.0-xsi)*(1.0-eta));
    m.set(5,0,0.5*(1.0+xsi)*(1.0+eta)*(1.0-eta));
    m.set(6,0,0.5*(1.0+xsi)*(1.0-xsi)*(1.0+eta));
    m.set(7,0,0.5*(1.0-xsi)*(1.0+eta)*(1.0-eta));
    m.set(0,1,0.25*(1.0-eta)*(2.0*xsi+eta));
    m.set(1,1,0.25*(1.0-eta)*(2.0*xsi-eta));
    m.set(2,1,0.25*(1.0+eta)*(2.0*xsi+eta));
    m.set(3,1,0.25*(1.0+eta)*(2.0*xsi-eta));
    m.set(4,1,-xsi*(1.0-eta));
    m.set(5,1,0.5*(1.0+eta)*(1.0-eta));
    m.set(6,1,-xsi*(1.0+eta));
    m.set(7,1,-0.5*(1.0+eta)*(1.0-eta));
    m.set(0,2,0.25*(1.0-xsi)*(xsi+2.0*eta));
    m.set(1,2,0.25*(1.0+xsi)*(-xsi+2.0*eta));
    m.set(2,2,0.25*(1.0+xsi)*(xsi+2.0*eta));
    m.set(3,2,0.25*(1.0-xsi)*(-xsi+2.0*eta));
    m.set(4,2,-0.5*(1.0+xsi)*(1.0-xsi));
    m.set(5,2,-(1.0+xsi)*eta);
    m.set(6,2,0.5*(1.0+xsi)*(1.0-xsi));
    m.set(7,2,-(1.0-xsi)*eta);
    return m;
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub> dS</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return `֐}gbNX
 */
  public Matrix shapeFunctionMatrix(Point3D p[],double coef){
    Matrix s=new DefaultMatrix(NODE_COUNT);
    for(int i=0;i<9;i++){		// KEXϕ
      int i1=i%3;
      int i2=(i/3);
      Matrix ss=shapeFunctionMatrix(p,GX3[i1],GX3[i2]);
      ss.mul(GW3[i1]*GW3[i2]);
      s.add(ss);
    }
    s.mul(coef);
    return s;
  }

//--------------------------------------------------------------------//
/**
 * `֐xNg<br>
 * { <i>N<sub>i</sub> dS</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return `֐xNg
 */
  public Vector shapeFunctionVector(Point3D p[],double coef){
    Vector s=new Vector(NODE_COUNT);
    for(int i=0;i<9;i++){		// KEXϕ
      int i1=i%3;
      int i2=(i/3);
      Vector ss=shapeFunctionVector(p,GX3[i1],GX3[i2]);
      ss.mul(GW3[i1]*GW3[i2]);
      s.add(ss);
    }
    s.mul(coef);
    return s;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @return rΏۂ̃IuWFNg <code>QuadangleBorder2</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof QuadangleBorder2){
      return super.equals(obj);
    }
    else{
      return false;
    }
  }
}
