package fem.component;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.text.DecimalFormat;
import java.util.Arrays;
import fem.AnalyzeType;
import fem.ElementBorder;
import fem.FEMDataModel;
import fem.MeshModel;
import fem.Node;
import fem.bc.BoundaryCondition;
import fem.bc.Load;
import fem.bc.Pressure;
import fem.bc.Restraint;
import fem.border.AbstractElementBorder;
import fem.result.ContourData;
import fem.result.Result;
import figure.Cube3D;
import figure.Point3D;
import figure.Transform;
import figure.Vector3D;
import gui.BasicViewCanvas;
import gui.ColorBar;

//--------------------------------------------------------------------//
/**
 * FEM f\LoX
 */
public class FEMViewCanvas extends BasicViewCanvas{

/** ̍ŏl */
  public static final double MIN_VALUE=1.0E-20;

/** l */
  protected static final DecimalFormat
    DFORM=new DecimalFormat("0.0000E00");

/** FEM f[^f */
  protected FEMDataModel model;

/** f\̈TCY */
  protected int viewSize;

/** ʒ_ */
  protected Point centerPoint;

/** G[bZ[W */
  protected String message;

/** tHg */
  protected Font font;

/** wiF */
  protected Color bgColor;

/** OiF */
  protected Color fgColor;

/** vf\F */
  protected Color elemColor;

/** S\F */
  protected Color restColor;

/** ׏d\F */
  protected Color loadColor;

/** ʈ\F */
  protected Color pressColor;

/** fsړW */
  protected double transCoef;

/** f]xW */
  protected double rotateCoef;

/** fgEkW */
  protected double zoomCoef;

/** ϊs */
  protected Transform transform;

/** J[o[ */
  protected ColorBar colorbar;

/** J[o[\̈\ <code>Rectangle</code> */
  protected Rectangle colorbarBound;

/** R^[}f[^ */
  protected ContourData contour;

/** ψʕ\{ */
  protected double dispCoef;

/** Z obt@ */
  protected double zbuffer[][];

/**
 * FEM f\LoX\z
 * @param model FEM f[^f
 */
  public FEMViewCanvas(FEMDataModel model,String message,Font font,
      	      	       Color bgColor,Color fgColor){
    this.model=model;
    this.message=message;
    this.font=font;
    setColors(bgColor,fgColor);
    elemColor=Color.GRAY;
    restColor=Color.BLUE;
    loadColor=Color.GREEN;
    pressColor=Color.MAGENTA;
    transform=new Transform();
    colorbar=null;
    contour=null;
    dispCoef=0.0;
  }

//--------------------------------------------------------------------//
/**
 * \Fݒ肷
 * @param bgColor wiF
 * @param fgColor OiF
 */
  protected void setColors(Color bgColor,Color fgColor){
    if(bgColor==null){
      this.bgColor=Color.WHITE;
    }
    else{
      this.bgColor=bgColor;
    }
    if(fgColor==null){
      this.fgColor=Color.BLACK;
    }
    else{
      this.fgColor=fgColor;
    }
    setBackground(bgColor);
  }

//--------------------------------------------------------------------//
/**
 * FEM f[^fԂ
 * @return FEM f[^f
 */
  public FEMDataModel getModel(){
    return model;
  }

//--------------------------------------------------------------------//
/**
 * f𕽍sړ
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void translate(int dx,int dy){
    transform.translate(transCoef*(double)dx,-transCoef*(double)dy,0.0);
  }

//--------------------------------------------------------------------//
/**
 * f]
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void rotate(int dx,int dy){
    transform.rotate(rotateCoef*(double)dy,rotateCoef*(double)dx,0.0);
  }

//--------------------------------------------------------------------//
/**
 * fgEk
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void zoom(int dx,int dy){
    transform.zoom(Math.pow(2.0,zoomCoef*(double)(dx+dy)));
  }

//--------------------------------------------------------------------//
/**
 * R^[\ݒ肷
 * @param param f[^̎ނ\ int ^
 * @param component f[^̐\ int ^
 */
  public void setContour(int param,int component){
    if(param==Result.NONE){
      contour=null;
      colorbar=null;
    }
    else if(model.getResult().calculated()){
      contour=new ContourData();
      contour.setContour(model.getResult(),param,component);
      colorbar=new ColorBar.Index(contour.colorbar(),10);
    }
    repaint();
  }

//--------------------------------------------------------------------//
/**
 * vZʕ\
 */
  public void clear(){
    contour=null;
    colorbar=null;
    dispCoef=0.0;
    repaint();
  }

//--------------------------------------------------------------------//
/**
 * J[o[ݒ肷
 * @param colorbar J[o[
 */
  public void setColorBar(ColorBar colorbar){
    this.colorbar=colorbar;
  }

//--------------------------------------------------------------------//
/**
 * ʂɑ΂䗦ψʕ\{vZ
 * @param coef ʂɑ΂䗦
 * @return ψʕ\{
 */
  public double getDisplacementViewCoef(double coef){
    double dispMax=Math.max(model.getResult().dispMax(),MIN_VALUE);
    return coef*(double)viewSize/dispMax;
  }

//--------------------------------------------------------------------//
/**
 * ψʕ\{ݒ肷
 * @param dispCoef ψʕ\{
 */
  public void setDisplacementCoef(double dispCoef){
    this.dispCoef=dispCoef;
    repaint();
  }

//--------------------------------------------------------------------//
/**
 * _ʓɂ邩ǂʂ
 * @param x <i>x</i> W
 * @param y <i>y</i> W
 * @return _ʓɂꍇ true
 */
  protected boolean innerFrame(int x,int y){
    return ((0<=x) && (x<panelSize.width)
      	 && (0<=y) && (y<panelSize.height));
  }

//--------------------------------------------------------------------//
/**
 * ĕ`悷
 * @param g `悷OtBbNXReLXg
 */
  public void update(Graphics g){
    if(model!=null){
      paintModel(bg);
    }
    else{
      drawErrorMessage(bg,message);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * `悷
 * @param g `悷OtBbNXReLXg
 */
  public void paint(Graphics g){
    setPaint();
    if(model!=null){
      paintModel(bg);
    }
    else{
      drawErrorMessage(bg,message);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * ʐݒ
 */
  protected void setPaint(){
    super.setPaint();
    viewSize=Math.min(panelSize.width,panelSize.height);
    centerPoint=new Point(viewSize/2,viewSize/2);
    transCoef=1.0;
    rotateCoef=Math.PI/(double)viewSize;
    zoomCoef=1.0/(double)viewSize;
    bg.setFont(font);
    zbuffer=new double[panelSize.width][panelSize.height];
    if(panelSize.width>panelSize.height){
      colorbarBound=new Rectangle
      	(panelSize.height*11/10,panelSize.height/10,
      	 panelSize.width/20,panelSize.height*4/5);
    }
    else{
      colorbarBound=new Rectangle
      	(panelSize.width/10,panelSize.width*11/10,
      	 panelSize.width*4/5,panelSize.height/20);
    }
  }

//--------------------------------------------------------------------//
/**
 * `e
 * @param g `悷OtBbNXReLXg
 */
  protected void paintModel(Graphics2D g){
    g.setColor(bgColor);
    g.fillRect(0,0,panelSize.width,panelSize.height);
    if(!model.modelExist()) return;

    Cube3D bounds=model.getMesh().getBounds();
    Point3D centerPoint=bounds.getCenterLocation();
    double boundSize=Math.max(Math.max(bounds.width,bounds.height),
      	      	      	      bounds.depth);
    double vCoef=0.8*(double)viewSize/Math.max(boundSize,MIN_VALUE);

    for(int i=0;i<panelSize.width;i++){
      Arrays.fill(zbuffer[i],-Double.MAX_VALUE);
    }

    int nodeCount=model.getMesh().getNodeCount();
    ViewPoint3D vp1[]=new ViewPoint3D[nodeCount];
    ViewPoint3D vp2[]=new ViewPoint3D[nodeCount];
    g.setColor(fgColor);
    if(model.getResult().calculated()
    && (model.getAnalyzeType()==AnalyzeType.STRUCTURE)){
      for(int i=0;i<nodeCount;i++){
      	Vector3D disp=model.getResult().getDisplacement(i);
      	ViewPoint3D vp[]=getViewPoint(i,vCoef,centerPoint,dispCoef,
      	      	      	      	      disp);
      	vp1[i]=vp[0];
      	vp2[i]=vp[1];
      }
      String dispMax=DFORM.format(model.getResult().dispMax());
      g.drawString("Displacement: "+dispMax,10,20);
    }
    else{
      for(int i=0;i<nodeCount;i++){
      	ViewPoint3D vp=getViewPoint(i,vCoef,centerPoint);
      	vp1[i]=vp;
      	vp2[i]=vp;
      }
      if(model.getResult().calculated()){
      	g.drawString("Calculated",10,20);
      }
      else{
      	int elementCount=model.getMesh().getElementCount();
      	g.drawString("Node:    "+nodeCount,10,20);
      	g.drawString("Element: "+elementCount,10,40);
      }
    }

    g.setColor(restColor);
    drawRestraint(g,vp1);
    g.setColor(loadColor);
    drawLoad(g,vCoef,centerPoint);
    g.setColor(pressColor);
    drawPress(g,vCoef,centerPoint);

    if(contour!=null){
      drawElement(g,vp2,contour,colorbar);
      if(panelSize.width>panelSize.height){
      	colorbar.drawVerticalBar(g,colorbarBound,fgColor);
      }
      else{
      	colorbar.drawHorizontalBar(g,colorbarBound,fgColor);
      }
    }
    else{
      drawElement(g,vp2);
    }
  }

//--------------------------------------------------------------------//
/**
 * vf`悷
 * @param g `悷OtBbNXReLXg
 * @param vp ߓ_̉ʏ̓_i[z
 * @param contour R^[}f[^
 * @param colorbar J[o[
 */
  protected void drawElement(Graphics2D g,ViewPoint3D vp[],
      	      	      	     ContourData contour,ColorBar colorbar){
    MeshModel mesh=model.getMesh();
    int faceCount=mesh.getFaceCount();
    for(int i=0;i<faceCount;i++){
      ElementBorder faces[]=mesh.getFreeFace(i).splitBorder();
      for(int j=0;j<faces.length;j++){
      	int fv[]=faces[j].nodes();
      	int gx[]=new int[fv.length];
      	int gy[]=new int[fv.length];
      	double gz[]=new double[fv.length];
      	double value[]=new double[fv.length];
      	for(int k=0;k<fv.length;k++){
      	  gx[k]=vp[fv[k]].getX(centerPoint);
      	  gy[k]=vp[fv[k]].getY(centerPoint);
      	  gz[k]=vp[fv[k]].z;
      	  value[k]=contour.getContourData(fv[k]);
      	}
      	fillPolygon(g,fv.length,gx,gy,gz,value,colorbar);
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * vf`悷
 * @param g `悷OtBbNXReLXg
 * @param vp ߓ_̉ʏ̓_i[z
 */
  protected void drawElement(Graphics2D g,ViewPoint3D vp[]){
    g.setColor(fgColor);
    MeshModel mesh=model.getMesh();
    int faceCount=mesh.getFaceCount();
    for(int i=0;i<faceCount;i++){
      ElementBorder face=mesh.getFreeFace(i);
      int fv[]=face.cycleNodes();
      int gx[]=new int[fv.length];
      int gy[]=new int[fv.length];
      double gz[]=new double[fv.length];
      for(int k=0;k<fv.length;k++){
      	gx[k]=vp[fv[k]].getX(centerPoint);
      	gy[k]=vp[fv[k]].getY(centerPoint);
      	gz[k]=vp[fv[k]].z;
      }
      drawPolygon(g,fv.length,gx,gy,gz);
    }
  }

//--------------------------------------------------------------------//
/**
 * S`悷
 * @param g `悷OtBbNXReLXg
 * @param vp ߓ_̉ʏ̓_i[z
 */
  protected void drawRestraint(Graphics2D g,ViewPoint3D vp[]){
    BoundaryCondition bc=model.getBoundaryCondition();
    int restraintCount=bc.restraintCount();
    for(int i=0;i<restraintCount;i++){
      Restraint restraint=bc.getRestraint(i);
      Point g1=vp[restraint.label()].getPoint(centerPoint);
      if(innerFrame(g1.x,g1.y)
      && (vp[restraint.node].z>=zbuffer[g1.x][g1.y])){
      	g.fillOval(g1.x-2,g1.y-2,4,4);
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * ׏d`悷
 * @param g `悷OtBbNXReLXg
 * @param vCoef f̕\{
 * @param center f̒S\ <code>Point3D</code>
 */
  protected void drawLoad(Graphics2D g,double vCoef,Point3D center){
    BoundaryCondition bc=model.getBoundaryCondition();
    double coef2=0.1*(double)viewSize/Math.max(bc.loadMax(),MIN_VALUE);
    double vsize=0.02*(double)viewSize;
    int loadCount=bc.loadCount();
    for(int i=0;i<loadCount;i++){
      Load load=bc.getLoad(i);
      ViewPoint3D vp[]=getViewPoint(load.label(),vCoef,center,coef2,
      	      	      	      	    load);
      Point g1=vp[0].getPoint(centerPoint);
      Point g2=vp[1].getPoint(centerPoint);
      if(innerFrame(g1.x,g1.y) && (vp[0].z>=zbuffer[g1.x][g1.y])){
      	drawVector(g,g1,g2,vsize);
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * ʈ`悷
 * @param g `悷OtBbNXReLXg
 * @param vCoef f̕\{
 * @param center f̒S\ <code>Point3D</code>
 */
  protected void drawPress(Graphics2D g,double vCoef,Point3D center){
    BoundaryCondition bc=model.getBoundaryCondition();
    double coef2=0.1*(double)viewSize/Math.max(bc.pressMax(),MIN_VALUE);
    double vsize=0.02*(double)viewSize;
    MeshModel mesh=model.getMesh();
    int pressureCount=bc.pressureCount();
    for(int i=0;i<pressureCount;i++){
      Pressure press=bc.getPressure(i);
      ElementBorder border=press.getBorder(mesh.getElement(press.label()));
      Node p[]=mesh.nodes(border);
      Point3D c=border.center(p);
      Vector3D norm=AbstractElementBorder.normalVector(p);
      norm.mul(press.press);
      ViewPoint3D vp[]=getViewPoint(c,vCoef,center,coef2,norm);
      Point g1=vp[0].getPoint(centerPoint);
      Point g2=vp[1].getPoint(centerPoint);
      if(innerFrame(g1.x,g1.y) && (vp[0].z>=zbuffer[g1.x][g1.y])){
      	drawVector(g,g2,g1,vsize);
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * xNg`悷
 * @param g `悷OtBbNXReLXg
 * @param p1 xNg̎n_
 * @param p2 xNg̏I_
 * @param vsize [\TCY
 */
  protected void drawVector(Graphics2D g,Point p1,Point p2,
      	      	      	    double vsize){
    g.drawLine(p1.x,p1.y,p2.x,p2.y);
    double dx=(double)(p1.x-p2.x);
    double dy=(double)(p1.y-p2.y);
    double dd=Math.sqrt(dx*dx+dy*dy);
    if(dd>0.0) vsize/=dd;
    double cc=0.5*vsize*Math.sqrt(3.0);
    double cs=0.5*vsize;
    int x1=p2.x+(int)(cc*dx-cs*dy);
    int y1=p2.y+(int)(cs*dx+cc*dy);
    int x2=p2.x+(int)(cc*dx+cs*dy);
    int y2=p2.y+(int)(-cs*dx+cc*dy);
    g.drawLine(x1,y1,p2.x,p2.y);
    g.drawLine(x2,y2,p2.x,p2.y);
  }

//--------------------------------------------------------------------//
/**
 * p``悷
 * @param g `悷OtBbNXReLXg
 * @param n _̐
 * @param gx _ <i>x</i> Wi[z
 * @param gy _ <i>y</i> Wi[z
 * @param gz _ <i>z</i> Wi[z
 */
  protected void drawPolygon(Graphics2D g,int n,int gx[],int gy[],
      	      	      	     double gz[]){
    g.draw(new Polygon(gx,gy,n));
  }

//--------------------------------------------------------------------//
/**
 * p``悷
 * @param g `悷OtBbNXReLXg
 * @param n _̐
 * @param gx _ <i>x</i> Wi[z
 * @param gy _ <i>y</i> Wi[z
 * @param gz _ <i>z</i> Wi[z
 */
  protected void fillPolygon(Graphics2D g,int n,int gx[],int gy[],
      	      	      	     double gz[]){
    Polygon pl=new Polygon(gx,gy,n);
    Rectangle rect=pl.getBounds();
    int xMin=Math.max(0,rect.x);
    int xMax=Math.min(panelSize.width,rect.x+rect.width);
    int yMin=Math.max(0,rect.y);
    int yMax=Math.min(panelSize.height,rect.y+rect.height);
    if((xMin>=xMax) || (yMin>=yMax) || (xMax<0) || (yMax<0)) return;
    byte drawtype[][]=new byte[xMax-xMin+1][yMax-yMin+1];
    for(int i=xMin;i<=xMax;i++){
      for(int j=yMin;j<=yMax;j++){
      	if(pl.contains(i,j)) drawtype[i-xMin][j-yMin]=1;
      	else                 drawtype[i-xMin][j-yMin]=0;
      }
    }
    switch(n){
      case 3:
      	for(int i=xMin;i<xMax;i++){
      	  for(int j=yMin;j<yMax;j++){
      	    if(drawtype[i-xMin][j-yMin]==1){
      	      fillPolygon(g,i,j,gz,
      	      	      	  PolygonCoordinates.triangle(i,j,gx,gy));
      	    }
      	  }
      	}
      	break;
      case 4:
      	for(int i=xMin;i<xMax;i++){
      	  for(int j=yMin;j<yMax;j++){
      	    if(drawtype[i-xMin][j-yMin]==1){
      	      fillPolygon(g,i,j,gz,
      	      	      	  PolygonCoordinates.quadangle(i,j,gx,gy));
      	    }
      	  }
      	}
      	break;
    }
  }

//--------------------------------------------------------------------//
/**
 * R^[p``悷
 * @param g `悷OtBbNXReLXg
 * @param n _̐
 * @param gx _ <i>x</i> Wi[z
 * @param gy _ <i>y</i> Wi[z
 * @param gz _ <i>z</i> Wi[z
 * @param value _̒li[z
 * @param colorbar J[o[
 */
  protected void fillPolygon(Graphics2D g,int n,int gx[],int gy[],
      	      	      	     double gz[],double value[],
      	      	      	     ColorBar colorbar){
    Polygon pl=new Polygon(gx,gy,n);
    Rectangle rect=pl.getBounds();
    int xMin=Math.max(0,rect.x);
    int xMax=Math.min(panelSize.width,rect.x+rect.width);
    int yMin=Math.max(0,rect.y);
    int yMax=Math.min(panelSize.height,rect.y+rect.height);
    if((xMin>=xMax) || (yMin>=yMax) || (xMax<0) || (yMax<0)) return;
    byte drawtype[][]=new byte[xMax-xMin+1][yMax-yMin+1];
    for(int i=xMin;i<=xMax;i++){
      for(int j=yMin;j<=yMax;j++){
      	if(pl.contains(i,j)) drawtype[i-xMin][j-yMin]=1;
      	else                 drawtype[i-xMin][j-yMin]=0;
      }
    }
    switch(n){
      case 3:
      	for(int i=xMin;i<xMax;i++){
      	  for(int j=yMin;j<yMax;j++){
      	    if(drawtype[i-xMin][j-yMin]==1){
      	      fillPolygon(g,i,j,gz,value,
      	      	      	  PolygonCoordinates.triangle(i,j,gx,gy));
      	    }
      	  }
      	}
      	break;
      case 4:
      	for(int i=xMin;i<xMax;i++){
      	  for(int j=yMin;j<yMax;j++){
      	    if(drawtype[i-xMin][j-yMin]==1){
      	      fillPolygon(g,i,j,gz,value,
      	      	      	  PolygonCoordinates.quadangle(i,j,gx,gy));
      	    }
      	  }
      	}
      	break;
    }
  }

//--------------------------------------------------------------------//
/**
 * p``悷
 * @param g `悷OtBbNXReLXg
 * @param x <i>x</i> W
 * @param y <i>y</i> W
 * @param gz _ <i>z</i> Wi[z
 * @param s p`W
 */
  protected void fillPolygon(Graphics2D g,int x,int y,double gz[],
      	      	      	     PolygonCoordinates s){
    if(s==null) return;
    double pz=s.valueAt(gz);
    if(pz>=zbuffer[x][y]){
      zbuffer[x][y]=pz;
      g.setColor(elemColor);
      g.drawLine(x,y,x,y);
    }
  }

//--------------------------------------------------------------------//
/**
 * p``悷
 * @param g `悷OtBbNXReLXg
 * @param x <i>x</i> W
 * @param y <i>y</i> W
 * @param gz _ <i>z</i> Wi[z
 * @param value _̒li[z
 * @param s p`W
 */
  protected void fillPolygon(Graphics2D g,int x,int y,double gz[],
      	      	      	     double value[],PolygonCoordinates s){
    if(s==null) return;
    double pz=s.valueAt(gz);
    if(pz>=zbuffer[x][y]){
      zbuffer[x][y]=pz;
      g.setColor(colorbar.color(s.valueAt(value)));
      g.drawLine(x,y,x,y);
    }
  }

//--------------------------------------------------------------------//
/**
 * ʏ̓_Ԃ
 * @param index ߓ_̃CfbNX
 * @param vCoef f̕\{
 * @param center f̒S\ <code>Point3D</code>
 * @return ʏ̓_
 */
  protected ViewPoint3D getViewPoint(int index,double vCoef,
      	      	      	      	     Point3D center){
    try{
      Point3D p=model.getMesh().getNode(index);
      ViewPoint3D vp=new ViewPoint3D(vCoef*(p.x-center.x),
      	      	      	      	     vCoef*(p.y-center.y),
      	      	      	      	     vCoef*(p.z-center.z));
      transform.trans(vp);
      return vp;
    }
    catch(Exception ex){
      return null;
    }
  }

//--------------------------------------------------------------------//
/**
 * ʏ̓_Ԃ
 * @param index ߓ_̃CfbNX
 * @param vCoef f̕\{
 * @param center f̒S\ <code>Point3D</code>
 * @param coef2 xNg̕\{
 * @param vector ߓ_̃xNg
 * @return ʏ̓_i[z
 */
  protected ViewPoint3D[] getViewPoint(int index,double vCoef,
      	      	      	      	       Point3D center,double coef2,
      	      	      	      	       Vector3D vector){
    return getViewPoint(model.getMesh().getNode(index),vCoef,center,
      	      	      	coef2,vector);
  }

//--------------------------------------------------------------------//
/**
 * ʏ̓_Ԃ
 * @param p ߓ_
 * @param vCoef f̕\{
 * @param center f̒S\ <code>Point3D</code>
 * @param coef2 xNg̕\{
 * @param vector ߓ_̃xNg
 * @return ʏ̓_i[z
 */
  protected ViewPoint3D[] getViewPoint(Point3D p,double vCoef,
      	      	      	      	       Point3D center,double coef2,
      	      	      	      	       Vector3D vector){
    try{
      ViewPoint3D vp[]=new ViewPoint3D[2];
      vp[0]=new ViewPoint3D(vCoef*(p.x-center.x),
      	      	      	    vCoef*(p.y-center.y),
      	      	      	    vCoef*(p.z-center.z));
      vp[1]=new ViewPoint3D(vp[0].x+coef2*vector.x,
      	      	      	    vp[0].y+coef2*vector.y,
      	      	      	    vp[0].z+coef2*vector.z);
      transform.trans(vp[0]);
      transform.trans(vp[1]);
      return vp;
    }
    catch(Exception ex){
      return null;
    }
  }
}
