package fem.element;
import fem.Material;
import fem.Node;
import fem.ShellElement;
import fem.result.Strain;
import fem.result.Stress;
import figure.DefaultMatrix;
import figure.Matrix;
import figure.Point3D;
import figure.Tensor3D;
import figure.Vector;
import figure.Vector3D;
import figure.Vector3DR;

//--------------------------------------------------------------------//
/**
 * VFvf (MITC) ̒ۃX[p[NX
 */
public abstract class AbstractShellElement
  extends AbstractElement implements ShellElement{

/** VFp[^̃CfbNX */
  protected int param;

/**
 * vf\z
 * @param label vfx
 * @param material ޗ̃CfbNX
 * @param param VFp[^̃CfbNX
 */
  protected AbstractShellElement(int label,int material,int param){
    super(label,material);
    this.param=param;
  }

//--------------------------------------------------------------------//
/**
 * VFp[^̃CfbNXԂ
 * @return VFp[^̃CfbNX
 */
  public int getParam(){
    return param;
  }

//--------------------------------------------------------------------//
/**
 * VFp[^̃CfbNXݒ肷
 * @param param VFp[^̃CfbNX
 */
  public void setParam(int param){
    this.param=param;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @return vfE
 */
  public int borderCount(){
    return 2;
  }

//--------------------------------------------------------------------//
/**
 * vfߓ_̊pxԂ
 * @param p ߓ_i[z
 * @return vfߓ_̊pxi[z
 */
  public double[] angle(Point3D p[]){
    int nodeCount=nodeCount();
    double angle[]=new double[nodeCount];
    for(int i=0;i<nodeCount;i++){
      Vector3D v1=new Vector3D(p[i],p[(i+1)%nodeCount]);
      Vector3D v2=new Vector3D(p[i],p[(i+nodeCount-1)%nodeCount]);
      angle[i]=Vector3D.angle(v1,v2);
    }
    return angle;
  }

//--------------------------------------------------------------------//
/**
 * ]}gbNXԂ
 * @param p ߓ_i[z
 * @return ]}gbNX
 */
  public Matrix dirMatrix(Point3D p[]){
    Vector3D v3=normalVector(p);
    Vector3D v2=v3.crossProduct(new Vector3D(p[0],p[1]));
    v2.normalize();
    Vector3D v1=v2.crossProduct(v3);
    Matrix d=new DefaultMatrix(3);
    d.set(0,0,v1.x);
    d.set(0,1,v2.x);
    d.set(0,2,v3.x);
    d.set(1,0,v1.y);
    d.set(1,1,v2.y);
    d.set(1,2,v3.y);
    d.set(2,0,v1.z);
    d.set(2,1,v2.z);
    d.set(2,2,v3.z);
    return d;
  }

//--------------------------------------------------------------------//
/**
 * RrsԂ
 * @param p ߓ_i[z
 * @param v @xNg
 * @param sf `֐s
 * @param t vf̌
 * @return Rrs
 */
  protected Matrix jacobianMatrix(Point3D p[],Vector3D v,Matrix sf,
      	      	      	      	  double t){
    Matrix jacob=new DefaultMatrix(3);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      double dndx2=sf.get(i,1);
      double dndy2=sf.get(i,2);
      jacob.add(0,0,dndx2*p[i].x);
      jacob.add(0,1,dndx2*p[i].y);
      jacob.add(0,2,dndx2*p[i].z);
      jacob.add(1,0,dndy2*p[i].x);
      jacob.add(1,1,dndy2*p[i].y);
      jacob.add(1,2,dndy2*p[i].z);
    }
    jacob.add(2,0,0.5*t*v.x);
    jacob.add(2,1,0.5*t*v.y);
    jacob.add(2,2,0.5*t*v.z);
    return jacob;
  }

//--------------------------------------------------------------------//
/**
 * RrsԂ
 * @param p ߓ_i[z
 * @param v @xNgi[z
 * @param sf `֐s
 * @param t vf̌
 * @return Rrs
 */
  protected Matrix jacobianMatrix(Point3D p[],Vector3D v[],Matrix sf,
      	      	      	      	  double t){
    Matrix jacob=new DefaultMatrix(3);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      double dndx2=sf.get(i,1);
      double dndy2=sf.get(i,2);
      double dndz2=0.5*t*sf.get(i,0);
      jacob.add(0,0,dndx2*p[i].x);
      jacob.add(0,1,dndx2*p[i].y);
      jacob.add(0,2,dndx2*p[i].z);
      jacob.add(1,0,dndy2*p[i].x);
      jacob.add(1,1,dndy2*p[i].y);
      jacob.add(1,2,dndy2*p[i].z);
      jacob.add(2,0,dndz2*v[i].x);
      jacob.add(2,1,dndz2*v[i].y);
      jacob.add(2,2,dndz2*v[i].z);
    }
    return jacob;
  }

//--------------------------------------------------------------------//
/**
 * RrAԂ
 * @param ja Rrs
 * @return RrA
 */
  protected double jacobian(Matrix ja){
    double j11=ja.get(1,1)*ja.get(2,2)-ja.get(1,2)*ja.get(2,1);
    double j21=ja.get(1,2)*ja.get(2,0)-ja.get(1,0)*ja.get(2,2);
    double j31=ja.get(1,0)*ja.get(2,1)-ja.get(1,1)*ja.get(2,0);
    return ja.get(0,0)*j11+ja.get(0,1)*j21+ja.get(0,2)*j31;
  }

//--------------------------------------------------------------------//
/**
 * tRrsԂ
 * @param ja Rrs
 * @param jacobian RrA
 * @param d ]}gbNX
 * @return tRrs
 */
  protected Matrix jacobInv(Matrix ja,double jacobian,Matrix d){
    Matrix matrix=new DefaultMatrix(3,3);
    if(jacobian==0.0) return matrix;
    double jinv=1.0/jacobian;
    double j11=ja.get(1,1)*ja.get(2,2)-ja.get(1,2)*ja.get(2,1);
    double j21=ja.get(1,2)*ja.get(2,0)-ja.get(1,0)*ja.get(2,2);
    double j31=ja.get(1,0)*ja.get(2,1)-ja.get(1,1)*ja.get(2,0);
    double j12=ja.get(2,1)*ja.get(0,2)-ja.get(2,2)*ja.get(0,1);
    double j22=ja.get(2,2)*ja.get(0,0)-ja.get(2,0)*ja.get(0,2);
    double j32=ja.get(2,0)*ja.get(0,1)-ja.get(2,1)*ja.get(0,0);
    double j13=ja.get(0,1)*ja.get(1,2)-ja.get(0,2)*ja.get(1,1);
    double j23=ja.get(0,2)*ja.get(1,0)-ja.get(0,0)*ja.get(1,2);
    double j33=ja.get(0,0)*ja.get(1,1)-ja.get(0,1)*ja.get(1,0);
    matrix.set(0,0,jinv*(d.get(0,0)*j11+d.get(1,0)*j21+d.get(2,0)*j31));
    matrix.set(0,1,jinv*(d.get(0,0)*j12+d.get(1,0)*j22+d.get(2,0)*j32));
    matrix.set(1,0,jinv*(d.get(0,1)*j11+d.get(1,1)*j21+d.get(2,1)*j31));
    matrix.set(1,1,jinv*(d.get(0,1)*j12+d.get(1,1)*j22+d.get(2,1)*j32));
    matrix.set(2,0,jinv*(d.get(0,2)*j11+d.get(1,2)*j21+d.get(2,2)*j31));
    matrix.set(2,1,jinv*(d.get(0,2)*j12+d.get(1,2)*j22+d.get(2,2)*j32));
    matrix.set(2,2,jinv*(d.get(0,2)*j13+d.get(1,2)*j23+d.get(2,2)*j33));
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * ʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param material ޗ
 * @return ʃ}gbNX
 */
  public Matrix massMatrix(Point3D p[],Material material){
    return massMatrix(p,material.density,1.0);
  }

//--------------------------------------------------------------------//
/**
 * ʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param density ޗ̖x
 * @return ʃ}gbNX
 */
  public Matrix massMatrix(Point3D p[],double density){
    return massMatrix(p,density,1.0);
  }

//--------------------------------------------------------------------//
/**
 * ʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param material ޗ
 * @param t 
 * @return ʃ}gbNX
 */
  public Matrix massMatrix(Point3D p[],Material material,double t){
    return massMatrix(p,material.density,t);
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param material ޗ
 * @return }gbNX
 */
  public Matrix stiffnessMatrix(Point3D p[],Material material){
    return stiffnessMatrix(p,material.matrix(),1.0);
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param d1  - c}gbNX
 * @return }gbNX
 */
  public Matrix stiffnessMatrix(Point3D p[],Matrix d1){
    return stiffnessMatrix(p,d1,1.0);
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param material ޗ
 * @param t 
 * @return }gbNX
 */
  public Matrix stiffnessMatrix(Point3D p[],Material material,double t){
    return stiffnessMatrix(p,material.matrix(),t);
  }

//--------------------------------------------------------------------//
/**
 * `֐̌z<br>
 * { <i>dN<sub>i</sub></i> /<i>dx</i>
 *   <i>dN<sub>i</sub></i> /<i>dy</i>
 *   <i>dN<sub>i</sub></i> /<i>dz</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param t vf̌
 * @return `֐̌zi[z
 */
  protected double[][] grad(Point3D p[],double xsi,double eta,double t){
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,normalVector(p),sf,t);
    double jacobian=jacobian(ja);
    return grad(p,ja,jacobian,sf,t);
  }

//--------------------------------------------------------------------//
/**
 * `֐̌z<br>
 * { <i>dN<sub>i</sub></i> /<i>dx</i>
 *   <i>dN<sub>i</sub></i> /<i>dy</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param ja Rrs
 * @param jacobian RrA
 * @param sf `֐s
 * @param t vf̌
 * @return `֐̌zi[z
 */
  protected double[][] grad(Point3D p[],Matrix ja,double jacobian,
      	      	      	    Matrix sf,double t){
    int nodeCount=nodeCount();
    Matrix jinv=jacobInv(ja,jacobian,dirMatrix(p));
    double grad[][]=new double[nodeCount][3];
    for(int i=0;i<nodeCount;i++){
      double dndx2=sf.get(i,1);
      double dndy2=sf.get(i,2);
      double dndz2=0.5*t*sf.get(i,0);
      for(int j=0;j<3;j++){
      	grad[i][j]=jinv.get(j,0)*dndx2+jinv.get(j,1)*dndy2
      	      	  +jinv.get(j,2)*dndz2;
      }
    }
    return grad;
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub></i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param w d݌W
 * @param t vf̌
 * @return `֐}gbNX
 */
  protected Matrix shapeFunctionMatrix(Point3D p[],double xsi,
      	      	      	      	       double eta,double w,double t){
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,normalVector(p),sf,t);
    double jacobian=jacobian(ja);
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(nodeCount);
    double coef=w*Math.abs(jacobian);
    for(int i=0;i<nodeCount;i++){
      for(int j=0;j<nodeCount;j++){
      	matrix.set(i,j,coef*sf.get(i,0)*sf.get(j,0));
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub></i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param w d݌W
 * @param t vf̌
 * @return gU}gbNX
 */
  protected Matrix gradMatrix(Point3D p[],double xsi,double eta,
      	      	      	      double w,double t){
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,normalVector(p),sf,t);
    double jacobian=jacobian(ja);
    double grad[][]=grad(p,ja,jacobian,sf,t);
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(nodeCount);
    double coef=w*Math.abs(jacobian);
    for(int i=0;i<nodeCount;i++){
      for(int j=0;j<nodeCount;j++){
      	matrix.set(i,j,coef*(grad[i][0]*grad[j][0]
      	      	      	    +grad[i][1]*grad[j][1]
      	      	      	    +grad[i][2]*grad[j][2]));
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * c - ψʃ}gbNXԂ
 * @param sf `֐s
 * @param ja Rrs
 * @param jacobian RrA
 * @param d ]}gbNX
 * @return c - ψʃ}gbNX̐i[}gbNX
 */
  protected Matrix strainMatrix1(Matrix sf,Matrix ja,double jacobian,
      	      	      	      	 Matrix d){
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(4,nodeCount);
    Matrix ji=jacobInv(ja,jacobian,d);
    for(int i=0;i<nodeCount;i++){
      matrix.set(0,i,ji.get(0,0)*sf.get(i,1)+ji.get(0,1)*sf.get(i,2));
      matrix.set(1,i,ji.get(1,0)*sf.get(i,1)+ji.get(1,1)*sf.get(i,2));
      matrix.set(2,i,ji.get(2,0)*sf.get(i,1)+ji.get(2,1)*sf.get(i,2));
      matrix.set(3,i,ji.get(2,2)*sf.get(i,0));
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * c - ψʃ}gbNXԂ<br>
 * c͗vfʍWAψʂ͑S̍W
 * @param sf `֐s
 * @param ja Rrs
 * @param jacobian RrA
 * @param d ]}gbNX
 * @param dt t]}gbNX
 * @param t vf̌
 * @param zeta ߓ_ <i></i> W
 * @return c - ψʃ}gbNX
 */
  protected Matrix strainMatrix(Matrix sf,Matrix ja,double jacobian,
      	      	      	      	Matrix d,Matrix dt,double t,
      	      	      	      	double zeta){
    Matrix dd=new DefaultMatrix(6,6);
    for(int i=0;i<3;i++){
      for(int j=0;j<3;j++){
      	dd.set(i,j,dt.get(i,j));
      	dd.set(i+3,j+3,dt.get(i,j));
      }
    }
    Matrix b=strainMatrix1(sf,ja,jacobian,d);
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(5,6*nodeCount);
    double z=0.5*t*zeta;
    for(int i=0;i<nodeCount;i++){
      Matrix m=new DefaultMatrix(5,6);
      m.set(0,0,b.get(0,i));
      m.set(0,4,z*b.get(0,i));
      m.set(1,1,b.get(1,i));
      m.set(1,3,-z*b.get(1,i));
      m.set(2,0,b.get(1,i));
      m.set(2,1,b.get(0,i));
      m.set(2,3,-z*b.get(0,i));
      m.set(2,4,z*b.get(1,i));
      m.set(3,1,b.get(2,i));
      m.set(3,2,b.get(1,i));
      m.set(3,3,-0.5*t*b.get(3,i)-z*b.get(2,i));
      m.set(4,0,b.get(2,i));
      m.set(4,2,b.get(0,i));
      m.set(4,4,0.5*t*b.get(3,i)+z*b.get(2,i));
      m=DefaultMatrix.mul(m,dd);
      int ib=6*i;
      for(int i1=0;i1<5;i1++){
      	for(int j1=0;j1<6;j1++){
      	  matrix.set(i1,ib+j1,m.get(i1,j1));
      	}
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * vfcԂ
 * @param p ߓ_i[z
 * @param u ߓ_ψʂi[z
 * @param n vf̖@xNg
 * @param d ]}gbNX
 * @param dt t]}gbNX
 * @param t vf̌
 * @param xsi ߓ_ <i></i> W
 * @param eta ߓ_ <i></i> W
 * @param zeta ߓ_ <i></i> W
 * @return vfc̐i[xNg
 */
  protected Vector strain(Point3D p[],Vector3DR u[],Vector3D n,
      	      	      	  Matrix d,Matrix dt,double t,
      	      	      	  double xsi,double eta,double zeta){
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,n,sf,t);
    Matrix b=strainMatrix(sf,ja,jacobian(ja),d,dt,t,zeta);
    return toVector(u).mul(b);
  }

//--------------------------------------------------------------------//
/**
 * vfcԂ
 * @param p ߓ_i[z
 * @param u ߓ_ψʂi[z
 * @param xsi ߓ_ <i></i> Wi[z
 * @param eta ߓ_ <i></i> Wi[z
 * @param zeta co͂ʂ <i></i> W
 * @param t vf̌
 * @return vfci[z
 */
  protected Strain[] strain(Point3D p[],Vector3DR u[],double xsi[],
      	      	      	    double eta[],double zeta,double t){
    int nodeCount=nodeCount();
    Vector3D n=normalVector(p);
    Matrix d=dirMatrix(p);
    Matrix dt=d.transposition();
    Tensor3D rot=new Tensor3D(dt);
    Strain strain[]=new Strain[nodeCount];
    for(int i=0;i<nodeCount;i++){
      strain[i]=toStrain(strain(p,u,n,d,dt,t,xsi[i],eta[i],zeta));
      strain[i].rotate(rot);
    }
    return strain;
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param p ߓ_i[z
 * @param u ߓ_ψʂi[z
 * @param d1  - c}gbNX
 * @param xsi ߓ_ <i></i> Wi[z
 * @param eta ߓ_ <i></i> Wi[z
 * @param zeta co͂ʂ <i></i> W
 * @param t vf̌
 * @return vf͂i[z
 */
  protected Stress[] stress(Point3D p[],Vector3DR u[],Matrix d1,
      	      	      	    double xsi[],double eta[],double zeta,
      	      	      	    double t){
    int nodeCount=nodeCount();
    Vector3D n=normalVector(p);
    Matrix d=dirMatrix(p);
    Matrix dt=d.transposition();
    Tensor3D rot=new Tensor3D(dt);
    Stress stress[]=new Stress[nodeCount];
    for(int i=0;i<nodeCount;i++){
      stress[i]=toStress
      	(strain(p,u,n,d,dt,t,xsi[i],eta[i],zeta).mul(d1));
      stress[i].rotate(rot);
    }
    return stress;
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param p ߓ_i[z
 * @param u ߓ_ψʂi[z
 * @param material ޗ
 * @param zeta co͂ʂ <i></i> W
 * @param t vf̌
 * @return vf͂i[z
 */
  public Stress[] stress(Point3D p[],Vector3DR u[],Material material,
      	      	      	 double zeta,double t){
    return stress(p,u,material.matrix(),zeta,t);
  }

//--------------------------------------------------------------------//
/**
 * xNgcɕϊ
 * @param v c̐i[xNg
 * @return c
 */
  protected Strain toStrain(Vector v){
    return new Strain(v.get(0),v.get(1),0.0,0.5*v.get(2),0.5*v.get(3),
      	      	      0.5*v.get(4));
  }

//--------------------------------------------------------------------//
/**
 * xNg͂ɕϊ
 * @param v ͂̐i[xNg
 * @return 
 */
  protected Stress toStress(Vector v){
    return new Stress(v.get(0),v.get(1),0.0,v.get(2),v.get(3),v.get(4));
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @return ̃IuWFNg̕\
 */
  public String toString(){
    StringBuilder sb=new StringBuilder(getName());
    sb.append('\t');
    sb.append(label);
    sb.append('\t');
    sb.append(material);
    sb.append('\t');
    sb.append(param);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      sb.append('\t');
      sb.append(nodes[i]);
    }
    return sb.toString();
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @param material ޗx
 * @param p ߓ_i[z
 * @return ̃IuWFNg̕\
 */
  public String toString(int material,Node p[]){
    return toString(material,param,p);
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @param material ޗx
 * @param param VFp[^x
 * @param p ߓ_i[z
 * @return ̃IuWFNg̕\
 */
  public String toString(int material,int param,Node p[]){
    StringBuilder sb=new StringBuilder(getName());
    sb.append('\t');
    sb.append(label);
    sb.append('\t');
    sb.append(material);
    sb.append('\t');
    sb.append(param);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      sb.append('\t');
      sb.append(p[i].label());
    }
    return sb.toString();
  }
}
