package fem.element;
import fem.Material;
import fem.Node;
import fem.SolidElement;
import fem.result.Strain;
import fem.result.Stress;
import figure.DefaultMatrix;
import figure.Matrix;
import figure.Point3D;
import figure.Vector;
import figure.Vector3D;

//--------------------------------------------------------------------//
/**
 * \bhvf̒ۃX[p[NX
 */
public abstract class AbstractSolidElement
  extends AbstractElement implements SolidElement{

/**
 * vf\z
 * @param label vfx
 * @param material ޗ̃CfbNX
 */
  protected AbstractSolidElement(int label,int material){
    super(label,material);
  }

//--------------------------------------------------------------------//
/**
 * RrsԂ
 * @param p ߓ_i[z
 * @param sf `֐s
 * @return Rrs
 */
  protected Matrix jacobianMatrix(Point3D p[],Matrix sf){
    Matrix jacob=new DefaultMatrix(3);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      double dndx2=sf.get(i,1);
      double dndy2=sf.get(i,2);
      double dndz2=sf.get(i,3);
      jacob.add(0,0,dndx2*p[i].x);
      jacob.add(0,1,dndx2*p[i].y);
      jacob.add(0,2,dndx2*p[i].z);
      jacob.add(1,0,dndy2*p[i].x);
      jacob.add(1,1,dndy2*p[i].y);
      jacob.add(1,2,dndy2*p[i].z);
      jacob.add(2,0,dndz2*p[i].x);
      jacob.add(2,1,dndz2*p[i].y);
      jacob.add(2,2,dndz2*p[i].z);
    }
    return jacob;
  }

//--------------------------------------------------------------------//
/**
 * RrAԂ
 * @param ja Rrs
 * @return RrA
 */
  protected double jacobian(Matrix ja){
    double j11=ja.get(1,1)*ja.get(2,2)-ja.get(1,2)*ja.get(2,1);
    double j21=ja.get(1,2)*ja.get(2,0)-ja.get(1,0)*ja.get(2,2);
    double j31=ja.get(1,0)*ja.get(2,1)-ja.get(1,1)*ja.get(2,0);
    return ja.get(0,0)*j11+ja.get(0,1)*j21+ja.get(0,2)*j31;
  }

//--------------------------------------------------------------------//
/**
 * `֐̌z<br>
 * { <i>dN<sub>i</sub></i> /<i>dx</i>
 *   <i>dN<sub>i</sub></i> /<i>dy</i>
 *   <i>dN<sub>i</sub></i> /<i>dz</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param zeta <i></i> W
 * @return `֐̌zi[z
 */
  protected Vector3D[] grad(Point3D p[],double xsi,double eta,
      	      	      	    double zeta){
    Matrix sf=shapeFunction(xsi,eta,zeta);
    Matrix ja=jacobianMatrix(p,sf);
    double jacobian=jacobian(ja);
    return grad(p,ja,jacobian,sf);
  }

//--------------------------------------------------------------------//
/**
 * `֐̌z<br>
 * { <i>dN<sub>i</sub></i> /<i>dx</i>
 *   <i>dN<sub>i</sub></i> /<i>dy</i>
 *   <i>dN<sub>i</sub></i> /<i>dz</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param ja Rrs
 * @param jacobian RrA
 * @param sf `֐s
 * @return `֐̌zi[z
 */
  protected Vector3D[] grad(Point3D p[],Matrix ja,double jacobian,
      	      	      	    Matrix sf){
    int nodeCount=nodeCount();
    double jinv=1.0/jacobian;
    double j11=jinv*(ja.get(1,1)*ja.get(2,2)-ja.get(1,2)*ja.get(2,1));
    double j12=jinv*(ja.get(2,1)*ja.get(0,2)-ja.get(2,2)*ja.get(0,1));
    double j13=jinv*(ja.get(0,1)*ja.get(1,2)-ja.get(0,2)*ja.get(1,1));
    double j21=jinv*(ja.get(1,2)*ja.get(2,0)-ja.get(1,0)*ja.get(2,2));
    double j22=jinv*(ja.get(2,2)*ja.get(0,0)-ja.get(2,0)*ja.get(0,2));
    double j23=jinv*(ja.get(0,2)*ja.get(1,0)-ja.get(0,0)*ja.get(1,2));
    double j31=jinv*(ja.get(1,0)*ja.get(2,1)-ja.get(1,1)*ja.get(2,0));
    double j32=jinv*(ja.get(2,0)*ja.get(0,1)-ja.get(2,1)*ja.get(0,0));
    double j33=jinv*(ja.get(0,0)*ja.get(1,1)-ja.get(0,1)*ja.get(1,0));
    Vector3D grad[]=new Vector3D[nodeCount];
    for(int i=0;i<nodeCount;i++){
      double dndx2=sf.get(i,1);
      double dndy2=sf.get(i,2);
      double dndz2=sf.get(i,3);
      grad[i]=new Vector3D(j11*dndx2+j12*dndy2+j13*dndz2,
      	      	      	   j21*dndx2+j22*dndy2+j23*dndz2,
      	      	      	   j31*dndx2+j32*dndy2+j33*dndz2);
    }
    return grad;
  }

//--------------------------------------------------------------------//
/**
 * c - ψʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param zeta <i></i> W
 * @return c - ψʃ}gbNX
 */
  protected Matrix strainMatrix(Point3D p[],double xsi,double eta,
      	      	      	      	double zeta){
    Matrix sf=shapeFunction(xsi,eta,zeta);
    Matrix ja=jacobianMatrix(p,sf);
    double jacobian=jacobian(ja);
    return strainMatrix(grad(p,ja,jacobian,sf));
  }

//--------------------------------------------------------------------//
/**
 * c - ψʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param ja Rrs
 * @param jacobian RrA
 * @param sf `֐s
 * @return c - ψʃ}gbNX
 */
  protected Matrix strainMatrix(Point3D p[],Matrix ja,double jacobian,
      	      	      	      	Matrix sf){
    return strainMatrix(grad(p,ja,jacobian,sf));
  }

//--------------------------------------------------------------------//
/**
 * c - ψʃ}gbNXԂ
 * @param grad `֐̌zi[z
 * @return c - ψʃ}gbNX
 */
  protected Matrix strainMatrix(Vector3D grad[]){
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(6,3*nodeCount);
    for(int i=0;i<nodeCount;i++){
      int ib=3*i;
      matrix.set(0,ib,grad[i].x);
      matrix.set(1,ib+1,grad[i].y);
      matrix.set(2,ib+2,grad[i].z);
      matrix.set(3,ib,grad[i].y);
      matrix.set(3,ib+1,grad[i].x);
      matrix.set(4,ib+1,grad[i].z);
      matrix.set(4,ib+2,grad[i].y);
      matrix.set(5,ib,grad[i].z);
      matrix.set(5,ib+2,grad[i].x);
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param d  - c}gbNX
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param zeta <i></i> W
 * @param w d݌W
 * @return }gbNX
 */
  protected Matrix stiffnessMatrix(Point3D p[],Matrix d,double xsi,
      	      	      	      	   double eta,double zeta,double w){
    Matrix sf=shapeFunction(xsi,eta,zeta);
    Matrix ja=jacobianMatrix(p,sf);
    return stiffnessMatrix(p,d,ja,jacobian(ja),sf,w);
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param d  - c}gbNX
 * @param ja Rrs
 * @param jacobian RrA
 * @param sf `֐s
 * @param w d݌W
 * @return }gbNX
 */
  protected Matrix stiffnessMatrix(Point3D p[],Matrix d,Matrix ja,
      	      	      	      	   double jacobian,Matrix sf,double w){
    Matrix b=strainMatrix(p,ja,jacobian,sf);
    Matrix k=
      	DefaultMatrix.mul(DefaultMatrix.mul(b.transposition(),d),b);
    k.mul(w*Math.abs(jacobian));
    return k;
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param strain vfci[z
 * @param material ޗ
 * @return vf͂i[z
 */
  public Stress[] stress(Strain strain[],Material material){
    return stress(strain,material.matrix());
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param strain vfci[z
 * @param d1  - c}gbNX
 * @return vf͂i[z
 */
  public Stress[] stress(Strain strain[],Matrix d1){
    int nodeCount=nodeCount();
    Stress str[]=new Stress[nodeCount];
    for(int i=0;i<nodeCount;i++){
      str[i]=new Stress(strain[i].vector().mul(d1));
    }
    return str;
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param p ߓ_i[z
 * @param d ߓ_ψʂi[z
 * @param material ޗ
 * @return vf͂i[z
 */
  public Stress[] stress(Point3D p[],Vector3D d[],Material material){
    return stress(strain(p,d),material.matrix());
  }

//--------------------------------------------------------------------//
/**
 * vfcԂ
 * @param p ߓ_i[z
 * @param d ߓ_ψʂi[z
 * @param xsi ߓ_ <i></i> Wi[z
 * @param eta ߓ_ <i></i> Wi[z
 * @param zeta ߓ_ <i></i> Wi[z
 * @return vfci[z
 */
  protected Strain[] strain(Point3D p[],Vector3D d[],double xsi[],
      	      	      	    double eta[],double zeta[]){
    int nodeCount=nodeCount();
    Strain strain[]=new Strain[nodeCount];
    Vector v=toVector(d);
    for(int i=0;i<nodeCount;i++){
      Matrix sf=shapeFunction(xsi[i],eta[i],zeta[i]);
      Matrix jacob=jacobianMatrix(p,sf);
      double jacobian=jacobian(jacob);
      strain[i]=new Strain(v.mul(strainMatrix(p,jacob,jacobian,sf)));
    }
    return strain;
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param p ߓ_i[z
 * @param d ߓ_ψʂi[z
 * @param material ޗ
 * @param xsi ߓ_ <i></i> Wi[z
 * @param eta ߓ_ <i></i> Wi[z
 * @param zeta ߓ_ <i></i> Wi[z
 * @return vf͂i[z
 */
  protected Stress[] stress(Point3D p[],Vector3D d[],Material material,
      	      	      	    double xsi[],double eta[],double zeta[]){
    return stress(p,d,material.matrix(),xsi,eta,zeta);
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param p ߓ_i[z
 * @param d ߓ_ψʂi[z
 * @param d1  - c}gbNX
 * @param xsi ߓ_ <i></i> Wi[z
 * @param eta ߓ_ <i></i> Wi[z
 * @param zeta ߓ_ <i></i> Wi[z
 * @return vf͂i[z
 */
  protected Stress[] stress(Point3D p[],Vector3D d[],Matrix d1,
      	      	      	    double xsi[],double eta[],double zeta[]){
    int nodeCount=nodeCount();
    Stress stress[]=new Stress[nodeCount];
    Vector v=toVector(d);
    for(int i=0;i<nodeCount;i++){
      Matrix sf=shapeFunction(xsi[i],eta[i],zeta[i]);
      Matrix jacob=jacobianMatrix(p,sf);
      double jacobian=jacobian(jacob);
      stress[i]=new Stress
      	(v.mul(strainMatrix(p,jacob,jacobian,sf)).mul(d1));
    }
    return stress;
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub></i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param zeta <i></i> W
 * @param w d݌W
 * @return `֐}gbNX
 */
  protected Matrix shapeFunctionMatrix(Point3D p[],
      	      	      	      	       double xsi,double eta,
      	      	      	      	       double zeta,double w){
    Matrix sf=shapeFunction(xsi,eta,zeta);
    Matrix ja=jacobianMatrix(p,sf);
    double jacobian=jacobian(ja);
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(nodeCount);
    double coef=w*Math.abs(jacobian);
    for(int i=0;i<nodeCount;i++){
      for(int j=0;j<nodeCount;j++){
      	matrix.set(i,j,coef*sf.get(i,0)*sf.get(j,0));
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub></i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param zeta <i></i> W
 * @param w d݌W
 * @return gU}gbNX
 */
  protected Matrix gradMatrix(Point3D p[],double xsi,double eta,
      	      	      	      double zeta,double w){
    Matrix sf=shapeFunction(xsi,eta,zeta);
    Matrix ja=jacobianMatrix(p,sf);
    double jacobian=jacobian(ja);
    Vector3D grad[]=grad(p,ja,jacobian,sf);
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(nodeCount);
    double coef=w*Math.abs(jacobian);
    for(int i=0;i<nodeCount;i++){
      for(int j=0;j<nodeCount;j++){
      	matrix.set(i,j,coef*grad[i].dotProduct(grad[j]));
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @return ̃IuWFNg̕\
 */
  public String toString(){
    StringBuilder sb=new StringBuilder(getName());
    sb.append('\t');
    sb.append(label);
    sb.append('\t');
    sb.append(material);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      sb.append('\t');
      sb.append(nodes[i]);
    }
    return sb.toString();
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @param material ޗx
 * @param p ߓ_i[z
 * @return ̃IuWFNg̕\
 */
  public String toString(int material,Node p[]){
    StringBuilder sb=new StringBuilder(getName());
    sb.append('\t');
    sb.append(label);
    sb.append('\t');
    sb.append(material);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      sb.append('\t');
      sb.append(p[i].label());
    }
    return sb.toString();
  }
}
