package fem.element;
import fem.ElementBorder;
import fem.border.QuadangleBorder1;
import fem.result.Strain;
import fem.result.Stress;
import figure.DefaultMatrix;
import figure.Matrix;
import figure.Point3D;
import figure.Vector;
import figure.Vector3D;
import figure.Vector3DR;

//--------------------------------------------------------------------//
/**
 * lp` 1 vf (MITC4) NX
 */
public class QuadElement1 extends AbstractShellElement{

/** vf */
  public static final String ELEMENT_NAME="QuadElement1";

/** ߓ_ */
  public static final int NODE_COUNT=4;

/** ߓ_ <i></i> Wi[z */
  protected static final double
    XSI[]={-1.0,1.0,1.0,-1.0};

/** ߓ_ <i></i> Wi[z */
  protected static final double
    ETA[]={-1.0,-1.0,1.0,1.0};

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param param VFp[^̃CfbNX
 */
  public QuadElement1(int label,int material,int param,int p1,int p2,
      	      	      int p3,int p4){
    super(label,material,param);
    nodes=new int[4];
    nodes[0]=p1;
    nodes[1]=p2;
    nodes[2]=p3;
    nodes[3]=p4;
  }

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param param VFp[^̃CfbNX
 * @param p ߓ_ԍi[z
 */
  public QuadElement1(int label,int material,int param,int p[]){
    this(label,material,param,p[0],p[1],p[2],p[3]);
  }

//--------------------------------------------------------------------//
/**
 * vf̂Ԃ
 * @return vf
 */
  public String getName(){
    return ELEMENT_NAME;
  }

//--------------------------------------------------------------------//
/**
 * ߓ_Ԃ
 * @return ߓ_
 */
  public int nodeCount(){
    return NODE_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @param element vfx
 * @param index vfẼCfbNX
 * @return vfEi[z
 */
  public ElementBorder border(int element,int index){
    switch(index){
      default:
      	return null;
      case 0:
      	return new QuadangleBorder1(element,nodes[0],nodes[1],nodes[2],
      	      	      	      	    nodes[3]);
      case 1:
      	return new QuadangleBorder1(element,nodes[0],nodes[3],nodes[2],
      	      	      	      	    nodes[1]);
    }
  }

//--------------------------------------------------------------------//
/**
 * @xNgԂ
 * @param p _i[z
 * @return @xNg
 */
  public Vector3D normalVector(Point3D p[]){
    Vector3D norm=Vector3D.crossProduct
      (new Vector3D(p[0],p[2]),new Vector3D(p[1],p[3]));
    norm.normalize();
    return norm;
  }

//--------------------------------------------------------------------//
/**
 * `֐s<br>
 * { <i>N<sub>i</sub></i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i> }<br>
 * Ԃ
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @return `֐s
 */
  public Matrix shapeFunction(double xsi,double eta){
    Matrix m=new DefaultMatrix(NODE_COUNT,3);
    m.set(0,0,0.25*(1.0-xsi)*(1.0-eta));
    m.set(1,0,0.25*(1.0+xsi)*(1.0-eta));
    m.set(2,0,0.25*(1.0+xsi)*(1.0+eta));
    m.set(3,0,0.25*(1.0-xsi)*(1.0+eta));
    m.set(0,1,-0.25*(1.0-eta));
    m.set(1,1,0.25*(1.0-eta));
    m.set(2,1,0.25*(1.0+eta));
    m.set(3,1,-0.25*(1.0+eta));
    m.set(0,2,-0.25*(1.0-xsi));
    m.set(1,2,-0.25*(1.0+xsi));
    m.set(2,2,0.25*(1.0+xsi));
    m.set(3,2,0.25*(1.0-xsi));
    return m;
  }

//--------------------------------------------------------------------//
/**
 * ʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param density ޗ̖x
 * @param t 
 * @return ʃ}gbNX
 */
  public Matrix massMatrix(Point3D p[],double density,double t){
    Matrix m=new DefaultMatrix(3*NODE_COUNT);
    double dx=(p[2].y-p[0].y)*(p[3].z-p[1].z)
      	     -(p[3].y-p[1].y)*(p[2].z-p[0].z);
    double dy=(p[2].z-p[0].z)*(p[3].x-p[1].x)
      	     -(p[3].z-p[1].z)*(p[2].x-p[0].x);
    double dz=(p[2].x-p[0].x)*(p[3].y-p[1].y)
      	     -(p[3].x-p[1].x)*(p[2].y-p[0].y);
    double s=0.5*Math.abs(dx*dx+dy*dy+dz*dz);
    double ds=0.25*s*t*density;
    for(int i=0;i<3*NODE_COUNT;i++) m.set(i,i,ds);
    return m;
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param d1  - c}gbNX
 * @param t 
 * @return }gbNX
 */
  public Matrix stiffnessMatrix(Point3D p[],Matrix d1,double t){
    Matrix k=new DefaultMatrix(6*NODE_COUNT);
    Vector3D n=normalVector(p);
    for(int i=0;i<4;i++){               // KEXϕ
      Matrix ks=stiffnessMatrix(p,d1,n,GX2[i%2],GX2[i/2],t);
      k.add(ks);
    }
    return k;
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param d1  - c}gbNX
 * @param n vf̖@xNg
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param t vf̌
 * @return }gbNX
 */
  protected Matrix stiffnessMatrix(Point3D p[],Matrix d1,Vector3D n,
      	      	      	      	   double xsi,double eta,double t){
    Matrix d=dirMatrix(p);
    Matrix dt=d.transposition();
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,n,sf,t);
    double jacobian=jacobian(ja);
    Matrix bc0=strainMatrix1(sf,ja,jacobian,d);
    Matrix sf1=shapeFunction(xsi,0.0);
    Matrix ja1=jacobianMatrix(p,n,sf1,t);
    double jacobian1=jacobian(ja1);
    Matrix bc[]=new Matrix[2];
    bc[0]=strainMatrix1(sf1,ja1,jacobian1,d);
    Matrix sf2=shapeFunction(0.0,eta);
    Matrix ja2=jacobianMatrix(p,n,sf2,t);
    double jacobian2=jacobian(ja2);
    bc[1]=strainMatrix1(sf2,ja2,jacobian2,d);
    Matrix phy=new DefaultMatrix(3);
    phy.set(0,1,1.0);
    phy.set(1,0,-1.0);
    Matrix phyt=phy.transposition();
    Matrix k=new DefaultMatrix(6*NODE_COUNT);

    double tt6=t*t/6.0;
    double ce1=1.0E-3*t*t*d1.get(3,3);
    double ce2=-ce1/(double)(NODE_COUNT-1);
    for(int i=0;i<NODE_COUNT;i++){
      for(int j=0;j<NODE_COUNT;j++){
      	Matrix k1=new DefaultMatrix(3);
      	Matrix k2=new DefaultMatrix(3);
      	Matrix k3=new DefaultMatrix(3);
      	for(int j1=0;j1<2;j1++){
      	  for(int j2=0;j2<2;j2++){
      	    k1.add(j1,j2,bc0.get(j1,i)*d1.get(j1,j2)*bc0.get(j2,j)
      	      	      	+bc0.get(1-j1,i)*d1.get(2,2)*bc0.get(1-j2,j));
      	  }
      	  double dd=d1.get(4-j1,4-j1);
      	  k1.add(j1,j1,bc[1-j1].get(2,i)*dd*bc[1-j1].get(2,j));
      	  k1.set(j1,2,bc[1-j1].get(2,i)*dd*bc[j1].get(j1,j));
      	  k1.set(2,j1,bc[j1].get(j1,i)*dd*bc[1-j1].get(2,j));
      	  k2.set(j1,j1,bc[1-j1].get(2,i)*dd*bc[1-j1].get(3,j));
      	  k2.set(2,j1,bc[1-j1].get(j1,i)*dd*bc[1-j1].get(3,j));
      	  k3.set(j1,j1,bc[1-j1].get(3,i)*dd*bc[1-j1].get(2,j));
      	  k3.set(j1,2,bc[1-j1].get(3,i)*dd*bc[1-j1].get(j1,j));
      	}
      	k1.set(2,2,bc[0].get(1,i)*d1.get(3,3)*bc[0].get(1,j)
      	      	  +bc[1].get(0,i)*d1.get(4,4)*bc[1].get(0,j));
      	Matrix k4=(DefaultMatrix)k1.clone();
      	k4.add(0,0,3.0*bc[1].get(3,i)*d1.get(4,4)*bc[1].get(3,j));
      	k4.add(1,1,3.0*bc[0].get(3,i)*d1.get(3,3)*bc[0].get(3,j));

      	k2=DefaultMatrix.mul(k2,phy);
      	k3=DefaultMatrix.mul(phyt,k3);
      	k4=DefaultMatrix.mul(DefaultMatrix.mul(phyt,k4),phy);

      	if(i==j) k4.add(2,2,ce1);
      	else     k4.add(2,2,ce2);
      	k1=DefaultMatrix.mul(DefaultMatrix.mul(d,k1),dt);
      	k2=DefaultMatrix.mul(DefaultMatrix.mul(d,k2),dt);
      	k3=DefaultMatrix.mul(DefaultMatrix.mul(d,k3),dt);
      	k4=DefaultMatrix.mul(DefaultMatrix.mul(d,k4),dt);
      	k1.mul(2.0);
      	k2.mul(t);
      	k3.mul(t);
      	k4.mul(tt6);
      	int i0=6*i;
      	int j0=6*j;
      	for(int j1=0;j1<3;j1++){
      	  for(int j2=0;j2<3;j2++){
      	    k.set(i0+j1,j0+j2,k1.get(j1,j2));
      	    k.set(i0+j1,j0+3+j2,k2.get(j1,j2));
      	    k.set(i0+3+j1,j0+j2,k3.get(j1,j2));
      	    k.set(i0+3+j1,j0+3+j2,k4.get(j1,j2));
      	  }
      	}
      }
    }
    k.mul(Math.abs(jacobian));
    return k;
  }

//--------------------------------------------------------------------//
/**
 * vfcԂ
 * @param p ߓ_i[z
 * @param u ߓ_ψʂi[z
 * @param zeta co͂ʂ <i></i> W
 * @param t vf̌
 * @return vfci[z
 */
  public Strain[] strain(Point3D p[],Vector3DR u[],double zeta,
      	      	      	 double t){
    return strain(p,u,XSI,ETA,zeta,t);
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param p ߓ_i[z
 * @param u ߓ_ψʂi[z
 * @param d1  - c}gbNX
 * @param zeta co͂ʂ <i></i> W
 * @param t vf̌
 * @return vf͂i[z
 */
  public Stress[] stress(Point3D p[],Vector3DR u[],Matrix d1,
      	      	      	 double zeta,double t){
    return stress(p,u,d1,XSI,ETA,zeta,t);
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return `֐}gbNX
 */
  public Matrix shapeFunctionMatrix(Point3D p[],double coef){
    return shapeFunctionMatrix(p,coef,1.0);
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @param t vf̌
 * @return `֐}gbNX
 */
  public Matrix shapeFunctionMatrix(Point3D p[],double coef,double t){
    Matrix s=new DefaultMatrix(NODE_COUNT);
    for(int i=0;i<NODE_COUNT;i++){		// KEXϕ
      s.add(shapeFunctionMatrix(p,GX2[i%2],GX2[(i/2)%2],coef,t));
    }
    return s;
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return gU}gbNX
 */
  public Matrix gradMatrix(Point3D p[],double coef){
    return gradMatrix(p,coef,1.0);
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @param t vf̌
 * @return gU}gbNX
 */
  public Matrix gradMatrix(Point3D p[],double coef,double t){
    Matrix g=new DefaultMatrix(NODE_COUNT);
    for(int i=0;i<NODE_COUNT;i++){		// KEXϕ
      g.add(gradMatrix(p,GX2[i%2],GX2[(i/2)%2],coef,t));
    }
    return g;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @return rΏۂ̃IuWFNg <code>QuadElement1</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof QuadElement1){
      return super.equals(obj);
    }
    else{
      return false;
    }
  }
}
