package fem.element;
import fem.ElementBorder;
import fem.border.TriangleBorder1;
import fem.result.Strain;
import fem.result.Stress;
import figure.DefaultMatrix;
import figure.Matrix;
import figure.Point3D;
import figure.Vector;
import figure.Vector3D;

//--------------------------------------------------------------------//
/**
 * lʑ 1 vfNX
 */
public class TetraElement1 extends AbstractSolidElement{

/** vf */
  public static final String ELEMENT_NAME="TetraElement1";

/** ߓ_ */
  public static final int NODE_COUNT=4;

/** vfE */
  public static final int BORDER_COUNT=4;

/** RrA - ̐όW */
  protected static double V_COEF=1.0/6.0;

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 */
  public TetraElement1(int label,int material,
      	      	       int p1,int p2,int p3,int p4){
    super(label,material);
    nodes=new int[4];
    nodes[0]=p1;
    nodes[1]=p2;
    nodes[2]=p3;
    nodes[3]=p4;
  }

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param p ߓ_ԍi[z
 */
  public TetraElement1(int label,int material,int p[]){
    this(label,material,p[0],p[1],p[2],p[3]);
  }

//--------------------------------------------------------------------//
/**
 * vf̂Ԃ
 * @return vf
 */
  public String getName(){
    return ELEMENT_NAME;
  }

//--------------------------------------------------------------------//
/**
 * ߓ_Ԃ
 * @return ߓ_
 */
  public int nodeCount(){
    return NODE_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @return vfE
 */
  public int borderCount(){
    return BORDER_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @param element vfx
 * @param index vfẼCfbNX
 * @return vfEi[z
 */
  public ElementBorder border(int element,int index){
    switch(index){
      default:
      	return null;
      case 0:
      	return new TriangleBorder1(element,nodes[0],nodes[2],nodes[1]);
      case 1:
      	return new TriangleBorder1(element,nodes[0],nodes[1],nodes[3]);
      case 2:
      	return new TriangleBorder1(element,nodes[1],nodes[2],nodes[3]);
      case 3:
      	return new TriangleBorder1(element,nodes[2],nodes[0],nodes[3]);
    }
  }

//--------------------------------------------------------------------//
/**
 * vfߓ_̊pxԂ
 * @param p ߓ_i[z
 * @return vfߓ_̊pxi[z
 */
  public double[] angle(Point3D p[]){
    double angle[]=new double[NODE_COUNT];
    for(int i=0;i<NODE_COUNT;i++){
      angle[i]=Vector3D.angle(new Vector3D(p[i],p[(i+1)%NODE_COUNT]),
      	      	      	      new Vector3D(p[i],p[(i+2)%NODE_COUNT]),
      	      	      	      new Vector3D(p[i],p[(i+3)%NODE_COUNT]));
    }
    return angle;
  }

//--------------------------------------------------------------------//
/**
 * vf̑̐ςԂ
 * @param p ߓ_i[z
 * @return vf̑̐
 */
  public double volume(Point3D p[]){
    return V_COEF*jacobian(p);
  }

//--------------------------------------------------------------------//
/**
 * RrAԂ
 * @param p ߓ_i[z
 * @return RrA
 */
  public double jacobian(Point3D p[]){
    double j11=(p[2].y-p[0].y)*(p[3].z-p[0].z)
      	      -(p[3].y-p[0].y)*(p[2].z-p[0].z);
    double j21=(p[3].y-p[0].y)*(p[1].z-p[0].z)
      	      -(p[1].y-p[0].y)*(p[3].z-p[0].z);
    double j31=(p[1].y-p[0].y)*(p[2].z-p[0].z)
      	      -(p[2].y-p[0].y)*(p[1].z-p[0].z);
    return (p[1].x-p[0].x)*j11+(p[2].x-p[0].x)*j21+(p[3].x-p[0].x)*j31;
  }

//--------------------------------------------------------------------//
/**
 * `֐s<br>
 * { <i>N<sub>i</sub></i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i> }<br>
 * Ԃ
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param zeta <i></i> W
 * @return `֐s
 */
  public Matrix shapeFunction(double xsi,double eta,double zeta){
    Matrix m=new DefaultMatrix(NODE_COUNT,4);
    m.set(0,0,1.0-xsi-eta-zeta);
    m.set(1,0,xsi);
    m.set(2,0,eta);
    m.set(3,0,zeta);
    m.set(0,1,-1.0);
    m.set(1,1,1.0);
    m.set(0,2,-1.0);
    m.set(2,2,1.0);
    m.set(0,3,-1.0);
    m.set(3,3,1.0);
    return m;
  }

//--------------------------------------------------------------------//
/**
 * `֐̌z<br>
 * { <i>dN<sub>i</sub></i> /<i>dx</i>
 *   <i>dN<sub>i</sub></i> /<i>dy</i>
 *   <i>dN<sub>i</sub></i> /<i>dz</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param jacobian RrA
 * @return `֐̌zi[z
 */
  protected Vector3D[] grad(Point3D p[],double jacobian){
    double jinv=1.0/jacobian;
    Vector3D grad[]=new Vector3D[NODE_COUNT];
    for(int i1=0;i1<NODE_COUNT;i1++){
      jinv*=-1.0;
      int i2=(i1+1)%4;
      int i3=(i1+2)%4;
      int i4=(i1+3)%4;
      double dndx2=jinv*((p[i3].y-p[i2].y)*(p[i4].z-p[i2].z)
      	      	      	-(p[i4].y-p[i2].y)*(p[i3].z-p[i2].z));
      double dndy2=jinv*((p[i3].z-p[i2].z)*(p[i4].x-p[i2].x)
      	      	      	-(p[i4].z-p[i2].z)*(p[i3].x-p[i2].x));
      double dndz2=jinv*((p[i3].x-p[i2].x)*(p[i4].y-p[i2].y)
      	      	      	-(p[i4].x-p[i2].x)*(p[i3].y-p[i2].y));
      grad[i1]=new Vector3D(dndx2,dndy2,dndz2);
    }
    return grad;
  }

//--------------------------------------------------------------------//
/**
 * c - ψʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param jacobian RrA
 * @return c - ψʃ}gbNX
 */
  public Matrix strainMatrix(Point3D p[],double jacobian){
    return strainMatrix(grad(p,jacobian));
  }

//--------------------------------------------------------------------//
/**
 * ʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param density ޗ̖x
 * @return ʃ}gbNX
 */
  public Matrix massMatrix(Point3D p[],double density){
    int size=3*NODE_COUNT;
    Matrix m=new DefaultMatrix(size);
    double value=1.0;
    for(int i=0;i<size;i+=3){
      m.set(i,i,1.0);
      m.set(i+1,i+1,1.0);
      m.set(i+2,i+2,1.0);
      for(int j=i+3;j<size;j+=3){
      	m.set(i,j,0.5);
      	m.set(i+1,j+1,0.5);
      	m.set(i+2,j+2,0.5);
      	m.set(j,i,0.5);
      	m.set(j+1,i+1,0.5);
      	m.set(j+2,i+2,0.5);
      }
    }
    m.mul(density*jacobian(p)/60.0);
    return m;
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param d1  - c}gbNX
 * @return }gbNX
 */
  public Matrix stiffnessMatrix(Point3D p[],Matrix d1){
    double jacobian=jacobian(p);
    Matrix b=strainMatrix(p,jacobian);
    Matrix k=
      DefaultMatrix.mul(DefaultMatrix.mul(b.transposition(),d1),b);
    k.mul(V_COEF*Math.abs(jacobian));
    return k;
  }

//--------------------------------------------------------------------//
/**
 * vfcԂ
 * @param p ߓ_i[z
 * @param d ߓ_ψʂi[z
 * @return vfci[z
 */
  public Strain[] strain(Point3D p[],Vector3D d[]){
    Vector v=toVector(d);
    double jacobian=jacobian(p);
    Strain str=new Strain(v.mul(strainMatrix(p,jacobian)));
    return new Strain[]{str,str,str,str};
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param p ߓ_i[z
 * @param d ߓ_ψʂi[z
 * @param d1  - c}gbNX
 * @return vf͂i[z
 */
  public Stress[] stress(Point3D p[],Vector3D d[],Matrix d1){
    Vector v=toVector(d);
    double jacobian=jacobian(p);
    Vector v2=v.mul(strainMatrix(p,jacobian));
    Stress str=new Stress(v2.mul(d1));
    return new Stress[]{str,str,str,str};
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return `֐}gbNX
 */
  public Matrix shapeFunctionMatrix(Point3D p[],double coef){
    Matrix s=new DefaultMatrix(NODE_COUNT);
    double jacobian=jacobian(p);
    double value=0.1*V_COEF*coef*Math.abs(jacobian);
    for(int i=0;i<NODE_COUNT;i++){
      s.set(i,i,value);
      for(int j=i+1;j<NODE_COUNT;j++){
      	s.set(i,j,0.5*value);
      	s.set(j,i,0.5*value);
      }
    }
    return s;
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return gU}gbNX
 */
  public Matrix gradMatrix(Point3D p[],double coef){
    Matrix g=new DefaultMatrix(NODE_COUNT);
    double jacobian=jacobian(p);
    Vector3D grad[]=grad(p,jacobian);
    double coef2=V_COEF*coef*Math.abs(jacobian);
    for(int i=0;i<NODE_COUNT;i++){
      for(int j=0;j<NODE_COUNT;j++){
      	g.set(i,j,coef2*grad[i].dotProduct(grad[j]));
      }
    }
    return g;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @return rΏۂ̃IuWFNg <code>TetraElement1</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof TetraElement1){
      return super.equals(obj);
    }
    else{
      return false;
    }
  }
}
