package fem.element;
import fem.ElementBorder;
import fem.border.TriangleBorder1;
import fem.result.Strain;
import fem.result.Stress;
import figure.DefaultMatrix;
import figure.Matrix;
import figure.Point3D;
import figure.Vector;
import figure.Vector3D;
import figure.Vector3DR;

//--------------------------------------------------------------------//
/**
 * Op` 1 vf (VF) NX
 */
public class TriElement1 extends AbstractShellElement{

/** vf */
  public static final String ELEMENT_NAME="TriElement1";

/** ߓ_ */
  public static final int NODE_COUNT=3;

/** ߓ_ <i></i> Wi[z */
  protected static final double XSI[]={0.0,1.0,0.0};

/** ߓ_ <i></i> Wi[z */
  protected static final double ETA[]={0.0,0.0,1.0};

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param param VFp[^̃CfbNX
 */
  public TriElement1(int label,int material,int param,
      	      	     int p1,int p2,int p3){
    super(label,material,param);
    nodes=new int[3];
    nodes[0]=p1;
    nodes[1]=p2;
    nodes[2]=p3;
  }

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param param VFp[^̃CfbNX
 * @param p ߓ_ԍi[z
 */
  public TriElement1(int label,int material,int param,int p[]){
    this(label,material,param,p[0],p[1],p[2]);
  }

//--------------------------------------------------------------------//
/**
 * vf̂Ԃ
 * @return vf
 */
  public String getName(){
    return ELEMENT_NAME;
  }

//--------------------------------------------------------------------//
/**
 * ߓ_Ԃ
 * @return ߓ_
 */
  public int nodeCount(){
    return NODE_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @param element vfx
 * @param index vfẼCfbNX
 * @return vfEi[z
 */
  public ElementBorder border(int element,int index){
    switch(index){
      default:
      	return null;
      case 0:
      	return new TriangleBorder1(element,nodes[0],nodes[1],nodes[2]);
      case 1:
      	return new TriangleBorder1(element,nodes[0],nodes[2],nodes[1]);
    }
  }

//--------------------------------------------------------------------//
/**
 * @xNgԂ
 * @param p _i[z
 * @return @xNg
 */
  public Vector3D normalVector(Point3D p[]){
    Vector3D norm=Vector3D.crossProduct
      (new Vector3D(p[0],p[1]),new Vector3D(p[0],p[2]));
    norm.normalize();
    return norm;
  }

//--------------------------------------------------------------------//
/**
 * P̌`֐s<br>
 * { <i>N<sub>i</sub></i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i> }<br>
 * Ԃ
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @return `֐s
 */
  public Matrix shapeFunction(double xsi,double eta){
    Matrix m=new DefaultMatrix(NODE_COUNT,3);
    m.set(0,0,1.0-xsi-eta);
    m.set(1,0,xsi);
    m.set(2,0,eta);
    m.set(0,1,-1.0);
    m.set(1,1,1.0);
    m.set(0,2,-1.0);
    m.set(2,2,1.0);
    return m;
  }

//--------------------------------------------------------------------//
/**
 * Q̌`֐s<br>
 * { <i>N<sub>i</sub></i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i> }<br>
 * Ԃ
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @return `֐s
 */
  public Matrix shapeFunction2(double xsi,double eta){
    Matrix m=new DefaultMatrix(2*NODE_COUNT,3);
    double xe=1.0-xsi-eta;
    m.set(0,0,xe*(2.0*xe-1.0));
    m.set(1,0,xsi*(2.0*xsi-1.0));
    m.set(2,0,eta*(2.0*eta-1.0));
    m.set(3,0,4.0*xe*xsi);
    m.set(4,0,4.0*xsi*eta);
    m.set(5,0,4.0*xe*eta);
    m.set(0,1,1.0-4.0*xe);
    m.set(1,1,4.0*xsi-1.0);
    m.set(3,1,4.0*(xe-xsi));
    m.set(4,1,4.0*eta);
    m.set(5,1,-4.0*eta);
    m.set(0,2,1.0-4.0*xe);
    m.set(2,2,4.0*eta-1.0);
    m.set(3,2,-4.0*xsi);
    m.set(4,2,4.0*xsi);
    m.set(5,2,4.0*(xe-eta));
    return m;
  }

//--------------------------------------------------------------------//
/**
 * px̌`֐s<br>
 * { <i>H<sub>xi</sub></i>
 *   <i>H<sub>yi</sub></i>
 *   <i>dH<sub>xi</sub></i> /<i>d</i>
 *   <i>dH<sub>yi</sub></i> /<i>d</i>
 *   <i>dH<sub>xi</sub></i> /<i>d</i>
 *   <i>dH<sub>yi</sub></i> /<i>d</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param dd ]}gbNX
 * @param xsi <i></i> W
 * @param eta <i></i> W
 */
  public Matrix shapeFunction3(Point3D p[],Matrix dd,double xsi,
      	      	      	       double eta){
    Matrix m=new DefaultMatrix(3*NODE_COUNT,6);
    Matrix sf2=shapeFunction2(xsi,eta);
    Vector3D d12=new Vector3D(p[0],p[1]);
    Vector3D d23=new Vector3D(p[1],p[2]);
    Vector3D d31=new Vector3D(p[2],p[0]);
    double l[]={1.0/d12.magnitudeSq(),1.0/d23.magnitudeSq(),
      	      	1.0/d31.magnitudeSq()};
    double x[]={dd.get(0,0)*d12.x+dd.get(1,0)*d12.y+dd.get(2,0)*d12.z,
      	      	dd.get(0,0)*d23.x+dd.get(1,0)*d23.y+dd.get(2,0)*d23.z,
      	      	dd.get(0,0)*d31.x+dd.get(1,0)*d31.y+dd.get(2,0)*d31.z};
    double y[]={dd.get(0,1)*d12.x+dd.get(1,1)*d12.y+dd.get(2,1)*d12.z,
      	      	dd.get(0,1)*d23.x+dd.get(1,1)*d23.y+dd.get(2,1)*d23.z,
      	      	dd.get(0,1)*d31.x+dd.get(1,1)*d31.y+dd.get(2,1)*d31.z};
    double a[]={1.5*l[0]*y[0],1.5*l[1]*y[1],1.5*l[2]*y[2]};
    double b[]={-1.5*l[0]*x[0],-1.5*l[1]*x[1],-1.5*l[2]*x[2]};
    double c[]={0.75*l[0]*y[0]*y[0]-0.5,0.75*l[1]*y[1]*y[1]-0.5,
      	      	0.75*l[2]*y[2]*y[2]-0.5};
    double d[]={0.75*l[0]*x[0]*y[0],0.75*l[1]*x[1]*y[1],
      	      	0.75*l[2]*x[2]*y[2]};
    double e[]={0.25-0.75*l[0]*y[0]*y[0],0.25-0.75*l[1]*y[1]*y[1],
      	      	0.25-0.75*l[2]*y[2]*y[2]};
    for(int i=0;i<3;i++){
      int i1=(i+2)%3;
      int i3=3*i;
      for(int j=0;j<3;j++){
      	int j2=2*j;
      	m.set(i3,j2,a[i1]*sf2.get(3+i1,j)-a[i]*sf2.get(3+i,j));
      	m.set(i3,j2+1,b[i1]*sf2.get(3+i1,j)-b[i]*sf2.get(3+i,j));
      	m.set(i3+1,j2,sf2.get(i,j)-c[i1]*sf2.get(3+i1,j)
      	      	      -c[i]*sf2.get(3+i,j));
      	double dn=d[i1]*sf2.get(3+i1,j)+d[i]*sf2.get(3+i,j);
      	m.set(i3+1,j2+1,dn);
      	m.set(i3+2,j2,dn);
      	m.set(i3+2,j2+1,sf2.get(i,j)-e[i1]*sf2.get(3+i1,j)
      	      	      	-e[i]*sf2.get(3+i,j));
      }
    }
    return m;
  }

//--------------------------------------------------------------------//
/**
 * RrAԂ
 * @param p ߓ_i[z
 * @return RrA
 */
  public double jacobian(Point3D p[]){
    double j1=(p[1].y-p[0].y)*(p[2].z-p[0].z)
      	     -(p[1].z-p[0].z)*(p[2].y-p[0].y);
    double j2=(p[1].z-p[0].z)*(p[2].x-p[0].x)
      	     -(p[1].x-p[0].x)*(p[2].z-p[0].z);
    double j3=(p[1].x-p[0].x)*(p[2].y-p[0].y)
      	     -(p[1].y-p[0].y)*(p[2].x-p[0].x);
    return Math.sqrt(j1*j1+j2*j2+j3*j3);
  }

//--------------------------------------------------------------------//
/**
 * ʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param density ޗ̖x
 * @param t 
 * @return ʃ}gbNX
 */
  public Matrix massMatrix(Point3D p[],double density,double t){
    Matrix m=new DefaultMatrix(3*NODE_COUNT);
    double dm=C1_3*0.5*jacobian(p)*t*density;
    for(int i=0;i<3*NODE_COUNT;i++) m.set(i,i,dm);
    return m;
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param d1  - c}gbNX
 * @param t 
 * @return }gbNX
 */
  public Matrix stiffnessMatrix(Point3D p[],Matrix d1,double t){
    Matrix d=dirMatrix(p);
    Matrix dt=d.transposition();
    Vector3D n=normalVector(p);

    Matrix sf1=shapeFunction(C1_3,C1_3);
    Matrix ja1=jacobianMatrix(p,n,sf1,t);
    double jac1=jacobian(ja1);
    Matrix jinv=jacobInv(ja1,jac1,d);
    Matrix b1=strainMatrix1(sf1,jinv);
    Matrix k1=DefaultMatrix.mul
      (DefaultMatrix.mul(b1.transposition(),d1),b1);
    k1.mul(Math.abs(jac1));

    Matrix k2=new DefaultMatrix(3*NODE_COUNT);
    for(int kk=0;kk<3;kk++){
      Matrix sf3=shapeFunction3(p,d,GTRI2[(int)XSI[kk]],
      	      	      	      	GTRI2[(int)ETA[kk]]);
      Matrix b2=strainMatrix2(sf3,jinv);
      k2.add(DefaultMatrix.mul
      	(DefaultMatrix.mul(b2.transposition(),d1),b2));
    }
    k2.mul(t*t*Math.abs(jac1)/36.0);
    double ce1=1.0E-3*t*t*Math.abs(jac1)*d1.get(2,2);
    double ce2=-ce1/(double)(NODE_COUNT-1);

    Matrix k=new DefaultMatrix(6*NODE_COUNT);
    Matrix dir=new DefaultMatrix(6*NODE_COUNT);
    for(int i=0;i<NODE_COUNT;i++){
      int i2=2*i;
      int i3=3*i;
      int i6=6*i;
      for(int j=0;j<NODE_COUNT;j++){
      	int j2=2*j;
      	int j3=3*j;
      	int j6=6*j;
      	k.set(i6,j6,k1.get(i2,j2));
      	k.set(i6,j6+1,k1.get(i2,j2+1));
      	k.set(i6+1,j6,k1.get(i2+1,j2));
      	k.set(i6+1,j6+1,k1.get(i2+1,j2+1));
      	for(int ii=0;ii<3;ii++){
      	  for(int jj=0;jj<3;jj++){
      	    k.set(i6+2+ii,j6+2+jj,k2.get(i3+ii,j3+jj));
      	  }
      	}
      	if(i==j) k.set(i6+5,j6+5,ce1);
      	else     k.set(i6+5,j6+5,ce2);
      }
      for(int ii=0;ii<3;ii++){
      	for(int jj=0;jj<3;jj++){
      	  dir.set(i6+ii,i6+jj,d.get(ii,jj));
      	  dir.set(i6+3+ii,i6+3+jj,d.get(ii,jj));
      	}
      }
    }
    k=DefaultMatrix.mul(DefaultMatrix.mul(dir,k),dir.transposition());
    return k;
  }

//--------------------------------------------------------------------//
/**
 * ʓc - ψʃ}gbNXԂ
 * @param sf `֐s
 * @param jinv tRrs
 * @return ʓc - ψʃ}gbNX
 */
  protected Matrix strainMatrix1(Matrix sf,Matrix jinv){
    Matrix b=new DefaultMatrix(3,2*NODE_COUNT);
    for(int i=0;i<NODE_COUNT;i++){
      double dndx=jinv.get(0,0)*sf.get(i,1)+jinv.get(0,1)*sf.get(i,2);
      double dndy=jinv.get(1,0)*sf.get(i,1)+jinv.get(1,1)*sf.get(i,2);
      int i2=2*i;
      b.set(0,i2,dndx);
      b.set(1,i2+1,dndy);
      b.set(2,i2,dndy);
      b.set(2,i2+1,dndx);
    }
    return b;
  }

//--------------------------------------------------------------------//
/**
 * ʊOc - ψʃ}gbNXԂ
 * @param sf `֐s
 * @param jinv tRrs
 * @return ʊOc - ψʃ}gbNX
 */
  protected Matrix strainMatrix2(Matrix sf,Matrix jinv){
    Matrix b=new DefaultMatrix(3,3*NODE_COUNT);
    for(int i=0;i<3*NODE_COUNT;i++){
      double hxx=jinv.get(0,0)*sf.get(i,2)+jinv.get(0,1)*sf.get(i,4);
      double hxy=jinv.get(1,0)*sf.get(i,2)+jinv.get(1,1)*sf.get(i,4);
      double hyx=jinv.get(0,0)*sf.get(i,3)+jinv.get(0,1)*sf.get(i,5);
      double hyy=jinv.get(1,0)*sf.get(i,3)+jinv.get(1,1)*sf.get(i,5);
      b.set(0,i,hyx);
      b.set(1,i,-hxy);
      b.set(2,i,hyy-hxx);
    }
    return b;
  }

//--------------------------------------------------------------------//
/**
 * c - ψʃ}gbNXԂ<br>
 * c͗vfʍWAψʂ͑S̍W
 * @param sf1 ʓό`̌`֐s
 * @param sf3 ʊOό`̌`֐s
 * @param jinv tRrs
 * @param dt t]}gbNX
 * @param t vf̌
 * @param zeta ߓ_ <i></i> W
 * @return c - ψʃ}gbNX
 */
  protected Matrix strainMatrix(Matrix sf1,Matrix sf3,Matrix jinv,
      	      	      	      	Matrix dt,double t,double zeta){
    Matrix dd=new DefaultMatrix(6,6);
    for(int i=0;i<3;i++){
      for(int j=0;j<3;j++){
      	dd.set(i,j,dt.get(i,j));
      	dd.set(i+3,j+3,dt.get(i,j));
      }
    }
    Matrix b1=strainMatrix1(sf1,jinv);
    Matrix b2=strainMatrix2(sf3,jinv);
    Matrix matrix=new DefaultMatrix(3,6*NODE_COUNT);
    double z=0.5*t*zeta;
    for(int i=0;i<NODE_COUNT;i++){
      int i2=2*i;
      int i3=3*i;
      int i6=6*i;
      Matrix m=new DefaultMatrix(3,6);
      for(int j=0;j<3;j++){
      	m.set(j,0,b1.get(j,i2));
      	m.set(j,1,b1.get(j,i2+1));
      	m.set(j,2,z*b2.get(j,i3));
      	m.set(j,3,z*b2.get(j,i3+1));
      	m.set(j,4,z*b2.get(j,i3+2));
      }
      m=DefaultMatrix.mul(m,dd);
      for(int i1=0;i1<3;i1++){
      	for(int j1=0;j1<6;j1++){
      	  matrix.set(i1,i6+j1,m.get(i1,j1));
      	}
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * vfcԂ
 * @param p ߓ_i[z
 * @param u ߓ_ψʂi[z
 * @param n vf̖@xNg
 * @param d ]}gbNX
 * @param dt t]}gbNX
 * @param t vf̌
 * @param xsi ߓ_ <i></i> W
 * @param eta ߓ_ <i></i> W
 * @param zeta ߓ_ <i></i> W
 * @return vfc̐i[xNg
 */
  protected Vector strain(Point3D p[],Vector3DR u[],Vector3D n,
      	      	      	  Matrix d,Matrix dt,double t,
      	      	      	  double xsi,double eta,double zeta){
    Matrix sf1=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,n,sf1,t);
    Matrix jinv=jacobInv(ja,jacobian(ja),d);
    Matrix sf3=shapeFunction3(p,d,xsi,eta);
    Matrix b=strainMatrix(sf1,sf3,jinv,dt,t,zeta);
    return toVector(u).mul(b);
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return `֐}gbNX
 */
  public Matrix shapeFunctionMatrix(Point3D p[],double coef){
    Matrix s=new DefaultMatrix(NODE_COUNT);
    double value=coef*jacobian(p)/12.0;
    for(int i=0;i<NODE_COUNT;i++){		// KEXϕ
      s.set(i,i,value);
      for(int j=i+1;j<NODE_COUNT;j++){
      	s.set(i,j,0.5*value);
      	s.set(j,i,0.5*value);
      }
    }
    return s;
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return gU}gbNX
 */
  public Matrix gradMatrix(Point3D p[],double coef){
    return gradMatrix(p,coef,1.0);
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @param t vf̌
 * @return gU}gbNX
 */
  public Matrix gradMatrix(Point3D p[],double coef,double t){
    return gradMatrix(p,C1_3,C1_3,coef,t);
  }

//--------------------------------------------------------------------//
/**
 * xNgcɕϊ
 * @param v c̐i[xNg
 * @return c
 */
  protected Strain toStrain(Vector v){
    return new Strain(v.get(0),v.get(1),0.0,0.5*v.get(2),0.0,0.0);
  }

//--------------------------------------------------------------------//
/**
 * xNg͂ɕϊ
 * @param v ͂̐i[xNg
 * @return 
 */
  protected Stress toStress(Vector v){
    return new Stress(v.get(0),v.get(1),0.0,v.get(2),0.0,0.0);
  }

//--------------------------------------------------------------------//
/**
 * vfcԂ
 * @param p ߓ_i[z
 * @param u ߓ_ψʂi[z
 * @param zeta co͂ʂ <i></i> W
 * @param t vf̌
 * @return vfci[z
 */
  public Strain[] strain(Point3D p[],Vector3DR u[],double zeta,
      	      	      	 double t){
    return strain(p,u,XSI,ETA,zeta,t);
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param p ߓ_i[z
 * @param u ߓ_ψʂi[z
 * @param d1  - c}gbNX
 * @param zeta co͂ʂ <i></i> W
 * @param t vf̌
 * @return vf͂i[z
 */
  public Stress[] stress(Point3D p[],Vector3DR u[],Matrix d1,
      	      	      	 double zeta,double t){
    return stress(p,u,d1,XSI,ETA,zeta,t);
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @return rΏۂ̃IuWFNg <code>TriElement1</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof TriElement1){
      return super.equals(obj);
    }
    else{
      return false;
    }
  }
}
