package fem.element;
import fem.ElementBorder;
import fem.border.QuadangleBorder1;
import fem.border.TriangleBorder1;
import fem.result.Strain;
import fem.result.Stress;
import figure.DefaultMatrix;
import figure.Matrix;
import figure.Point3D;
import figure.Vector;
import figure.Vector3D;

//--------------------------------------------------------------------//
/**
 * ` 1 vfNX
 */
public class WedgeElement1 extends AbstractSolidElement{

/** vf */
  public static final String ELEMENT_NAME="WedgeElement1";

/** ߓ_ */
  public static final int NODE_COUNT=6;

/** vfE */
  public static final int BORDER_COUNT=5;

/** ߓ_ <i></i> Wi[z */
  protected static final double XSI[]={0.0,1.0,0.0,0.0,1.0,0.0};

/** ߓ_ <i></i> Wi[z */
  protected static final double ETA[]={0.0,0.0,1.0,0.0,0.0,1.0};

/** ߓ_ <i></i> Wi[z */
  protected static final double ZETA[]={-1.0,-1.0,-1.0,1.0,1.0,1.0};

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 */
  public WedgeElement1(int label,int material,int p1,int p2,int p3,
      	      	       int p4,int p5,int p6){
    super(label,material);
    nodes=new int[6];
    nodes[0]=p1;
    nodes[1]=p2;
    nodes[2]=p3;
    nodes[3]=p4;
    nodes[4]=p5;
    nodes[5]=p6;
  }

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param p ߓ_ԍi[z
 */
  public WedgeElement1(int label,int material,int p[]){
    this(label,material,p[0],p[1],p[2],p[3],p[4],p[5]);
  }

//--------------------------------------------------------------------//
/**
 * vf̂Ԃ
 * @return vf
 */
  public String getName(){
    return ELEMENT_NAME;
  }

//--------------------------------------------------------------------//
/**
 * ߓ_Ԃ
 * @return ߓ_
 */
  public int nodeCount(){
    return NODE_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @return vfE
 */
  public int borderCount(){
    return BORDER_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @param element vfx
 * @param index vfẼCfbNX
 * @return vfEi[z
 */
  public ElementBorder border(int element,int index){
    switch(index){
      default:
      	return null;
      case 0:
      	return new TriangleBorder1(element,nodes[0],nodes[2],nodes[1]);
      case 1:
      	return new TriangleBorder1(element,nodes[3],nodes[4],nodes[5]);
      case 2:
      	return new QuadangleBorder1(element,nodes[0],nodes[1],nodes[4],
      	      	      	      	    nodes[3]);
      case 3:
      	return new QuadangleBorder1(element,nodes[1],nodes[2],nodes[5],
      	      	      	      	    nodes[4]);
      case 4:
      	return new QuadangleBorder1(element,nodes[2],nodes[0],nodes[3],
      	      	      	      	    nodes[5]);
    }
  }

//--------------------------------------------------------------------//
/**
 * vfߓ_̊pxԂ
 * @param p ߓ_i[z
 * @return vfߓ_̊pxi[z
 */
  public double[] angle(Point3D p[]){
    double angle[]=new double[NODE_COUNT];
    for(int i=0;i<3;i++){
      Vector3D v3=new Vector3D(p[i],p[i+3]);
      angle[i]=Vector3D.angle(new Vector3D(p[i],p[(i+1)%3]),
      	      	      	      new Vector3D(p[i],p[(i+2)%3]),v3);
      v3.mul(-1.0);
      angle[i+3]=Vector3D.angle(new Vector3D(p[i+3],p[(i+1)%3+3]),
      	      	      	      	new Vector3D(p[i+3],p[(i+2)%3+3]),v3);
    }
    return angle;
  }

//--------------------------------------------------------------------//
/**
 * `֐s<br>
 * { <i>N<sub>i</sub></i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i> }<br>
 * Ԃ
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param zeta <i></i> W
 * @return `֐s
 */
  public Matrix shapeFunction(double xsi,double eta,double zeta){
    Matrix m=new DefaultMatrix(NODE_COUNT,4);
    m.set(0,0,0.5*(1.0-xsi-eta)*(1.0-zeta));
    m.set(1,0,0.5*xsi*(1.0-zeta));
    m.set(2,0,0.5*eta*(1.0-zeta));
    m.set(3,0,0.5*(1.0-xsi-eta)*(1.0+zeta));
    m.set(4,0,0.5*xsi*(1.0+zeta));
    m.set(5,0,0.5*eta*(1.0+zeta));
    m.set(0,1,-0.5*(1.0-zeta));
    m.set(1,1,0.5*(1.0-zeta));
    m.set(3,1,-0.5*(1.0+zeta));
    m.set(4,1,0.5*(1.0+zeta));
    m.set(0,2,-0.5*(1.0-zeta));
    m.set(2,2,0.5*(1.0-zeta));
    m.set(3,2,-0.5*(1.0+zeta));
    m.set(5,2,0.5*(1.0+zeta));
    m.set(0,3,-0.5*(1.0-xsi-eta));
    m.set(1,3,-0.5*xsi);
    m.set(2,3,-0.5*eta);
    m.set(3,3,0.5*(1.0-xsi-eta));
    m.set(4,3,0.5*xsi);
    m.set(5,3,0.5*eta);
    return m;
  }

//--------------------------------------------------------------------//
/**
 * ʃ}gbNXԂ
 * @param p ߓ_i[z
 * @param density ޗ̖x
 * @return ʃ}gbNX
 */
  public Matrix massMatrix(Point3D p[],double density){
    Matrix m=new DefaultMatrix(3*NODE_COUNT);
    for(int i=0;i<NODE_COUNT;i++){
      for(int j=0;j<NODE_COUNT;j++){
      	double value=1.0;
      	if(i%3!=j%3) value*=0.5;
      	if(i/3!=j/3) value*=0.5;
      	int i3=3*i;
      	int j3=3*j;
      	m.set(i3,j3,value);
      	m.set(i3+1,j3+1,value);
      	m.set(i3+2,j3+2,value);
      }
    }
    double jacobian=0.0;
    for(int i=0;i<2;i++){              // KEXϕ
      Matrix sf=shapeFunction(C1_3,C1_3,GX2[i]);
      Matrix jacob=jacobianMatrix(p,sf);
      jacobian+=jacobian(jacob);
    }
    m.mul(density*jacobian/72.0);
    return m;
  }

//--------------------------------------------------------------------//
/**
 * }gbNXԂ
 * @param p ߓ_i[z
 * @param d1  - c}gbNX
 * @return }gbNX
 */
  public Matrix stiffnessMatrix(Point3D p[],Matrix d1){
    Matrix k=new DefaultMatrix(3*NODE_COUNT);
    for(int i=0;i<2;i++){              // KEXϕ
      k.add(stiffnessMatrix(p,d1,C1_3,C1_3,GX2[i],1.0));
    }
    k.mul(0.5);
    return k;
  }

//--------------------------------------------------------------------//
/**
 * vfcԂ
 * @param p ߓ_i[z
 * @param d ߓ_ψʂi[z
 * @return vfci[z
 */
  public Strain[] strain(Point3D p[],Vector3D d[]){
    return strain(p,d,XSI,ETA,ZETA);
  }

//--------------------------------------------------------------------//
/**
 * vf͂Ԃ
 * @param p ߓ_i[z
 * @param d ߓ_ψʂi[z
 * @param d1  - c}gbNX
 * @return vf͂i[z
 */
  public Stress[] stress(Point3D p[],Vector3D d[],Matrix d1){
    return stress(p,d,d1,XSI,ETA,ZETA);
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return `֐}gbNX
 */
  public Matrix shapeFunctionMatrix(Point3D p[],double coef){
    Matrix s=new DefaultMatrix(NODE_COUNT);
    for(int i=0;i<2;i++){		// KEXϕ
      s.add(shapeFunctionMatrix(p,C1_3,C1_3,GX2[i],0.5*coef));
    }
    return s;
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return gU}gbNX
 */
  public Matrix gradMatrix(Point3D p[],double coef){
    Matrix g=new DefaultMatrix(NODE_COUNT);
    for(int i=0;i<2;i++){		// KEXϕ
      g.add(gradMatrix(p,C1_3,C1_3,GX2[i],0.5*coef));
    }
    return g;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @return rΏۂ̃IuWFNg <code>WedgeElement1</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof WedgeElement1){
      return super.equals(obj);
    }
    else{
      return false;
    }
  }
}
