package fem.io;
import java.io.BufferedReader;
import java.util.ArrayList;
import figure.Vector3D;
import fem.Element;
import fem.FEMDataModel;
import fem.Material;
import fem.MeshModel;
import fem.Node;
import fem.ShellElement;
import fem.ShellParameter;
import fem.bc.HeatTransferBound;
import fem.bc.Load;
import fem.bc.Pressure;
import fem.bc.Restraint;
import fem.bc.Temperature;
import fem.element.HexaElement1;
import fem.element.HexaElement2;
import fem.element.QuadElement1;
import fem.element.TetraElement1;
import fem.element.TetraElement2;
import fem.element.TriElement1;
import fem.element.WedgeElement1;
import fem.element.WedgeElement2;
import fem.result.Result;
import fem.result.Strain;
import fem.result.Stress;
import figure.Vector3DR;
import fileio.TextDataReader;

//--------------------------------------------------------------------//
/**
 * FEM f[^ǂݍ݃NX
 */
public class FEMModelReader extends TextDataReader{

/** FEM f[^f */
  protected FEMDataModel model;

/** bVf */
  protected MeshModel mesh;

/**
 * FEM f[^ǂݍ݃IuWFNg\z
 */
  public FEMModelReader(FEMDataModel model){
    this.model=model;
    mesh=model.getMesh();
  }

//--------------------------------------------------------------------//
/**
 * FEM f[^ǂݍ
 * @param br ̓[_[
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  public void readData(BufferedReader br) throws Exception{
    int vertex[]=new int[20];
    model.clear();
    while(true){
      String line=br.readLine();
      if(line==null) break;
      String ss[]=split(line);
      if(ss.length>0){
      	String keyWord=ss[0];
      	if(keyWord.equals("Material")){
      	  int label=Integer.parseInt(ss[1]);
      	  double youngs=Double.parseDouble(ss[2]);
      	  double pois=Double.parseDouble(ss[3]);
      	  double share=Double.parseDouble(ss[4]);
      	  double dens=Double.parseDouble(ss[5]);
      	  double hcon=Double.parseDouble(ss[6]);
      	  double sh=Double.parseDouble(ss[7]);
      	  Material material=
      	    new Material(label,youngs,pois,dens,hcon,sh);
      	  material.matrix3D();
      	  model.addMaterial(material);
      	}
      	else if(keyWord.equals("ShellParameter")){
      	  int label=Integer.parseInt(ss[1]);
      	  double t=Double.parseDouble(ss[2]);
      	  model.addShellParameter(new ShellParameter(label,t));
      	}
      	else if(keyWord.equals("Node")){
      	  int label=Integer.parseInt(ss[1]);
      	  double x=Double.parseDouble(ss[2]);
      	  double y=Double.parseDouble(ss[3]);
      	  double z=Double.parseDouble(ss[4]);
      	  mesh.addNode(new Node(label,x,y,z));
      	}
      	else if(keyWord.equals(TriElement1.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  int param=Integer.parseInt(ss[3]);
      	  for(int i=0;i<3;i++){
      	    vertex[i]=Integer.parseInt(ss[4+i]);
      	  }
      	  mesh.addElement(new TriElement1(label,material,param,vertex));
      	}
      	else if(keyWord.equals(QuadElement1.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  int param=Integer.parseInt(ss[3]);
      	  for(int i=0;i<4;i++){
      	    vertex[i]=Integer.parseInt(ss[4+i]);
      	  }
      	  mesh.addElement(new QuadElement1(label,material,param,vertex));
      	}
      	else if(keyWord.equals(TetraElement1.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  for(int i=0;i<4;i++){
      	    vertex[i]=Integer.parseInt(ss[3+i]);
      	  }
      	  mesh.addElement(new TetraElement1(label,material,vertex));
      	}
      	else if(keyWord.equals(WedgeElement1.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  for(int i=0;i<6;i++){
      	    vertex[i]=Integer.parseInt(ss[3+i]);
      	  }
      	  mesh.addElement(new WedgeElement1(label,material,vertex));
      	}
      	else if(keyWord.equals(HexaElement1.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  for(int i=0;i<8;i++){
      	    vertex[i]=Integer.parseInt(ss[3+i]);
      	  }
      	  mesh.addElement(new HexaElement1(label,material,vertex));
      	}
      	else if(keyWord.equals(TetraElement2.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  for(int i=0;i<10;i++){
      	    vertex[i]=Integer.parseInt(ss[3+i]);
      	  }
      	  mesh.addElement(new TetraElement2(label,material,vertex));
      	}
      	else if(keyWord.equals(WedgeElement2.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  for(int i=0;i<15;i++){
      	    vertex[i]=Integer.parseInt(ss[3+i]);
      	  }
      	  mesh.addElement(new WedgeElement2(label,material,vertex));
      	}
      	else if(keyWord.equals(HexaElement2.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  for(int i=0;i<20;i++){
      	    vertex[i]=Integer.parseInt(ss[3+i]);
      	  }
      	  mesh.addElement(new HexaElement2(label,material,vertex));
      	}
      	else if(keyWord.equals("Restraint")){
      	  int label=Integer.parseInt(ss[1]);
      	  boolean rx=(Integer.parseInt(ss[2])!=0);
      	  double x=Double.parseDouble(ss[3]);
      	  boolean ry=(Integer.parseInt(ss[4])!=0);
      	  double y=Double.parseDouble(ss[5]);
      	  boolean rz=(Integer.parseInt(ss[6])!=0);
      	  double z=Double.parseDouble(ss[7]);
      	  if(ss.length<14){
      	    model.getBoundaryCondition().addRestraint
      	      (new Restraint(label,rx,ry,rz,x,y,z));
      	  }
      	  else{
      	    boolean rorx=(Integer.parseInt(ss[8])!=0);
      	    double rox=Double.parseDouble(ss[9]);
      	    boolean rory=(Integer.parseInt(ss[10])!=0);
      	    double roy=Double.parseDouble(ss[11]);
      	    boolean rorz=(Integer.parseInt(ss[12])!=0);
      	    double roz=Double.parseDouble(ss[13]);
      	    model.getBoundaryCondition().addRestraint
      	      (new Restraint(label,rx,ry,rz,rorx,rory,rorz,x,y,z,
      	      	      	     rox,roy,roz));
      	  }
      	}
      	else if(keyWord.equals("Load")){
      	  int label=Integer.parseInt(ss[1]);
      	  double x=Double.parseDouble(ss[2]);
      	  double y=Double.parseDouble(ss[3]);
      	  double z=Double.parseDouble(ss[4]);
      	  if(ss.length<8){
      	    model.getBoundaryCondition().addLoad(new Load(label,x,y,z));
      	  }
      	  else{
      	    double rx=Double.parseDouble(ss[5]);
      	    double ry=Double.parseDouble(ss[6]);
      	    double rz=Double.parseDouble(ss[7]);
      	    model.getBoundaryCondition().addLoad(new Load(label,x,y,z,
      	      	      	      	      	      	      	  rx,ry,rz));
      	  }
      	}
      	else if(keyWord.equals("Pressure")){
      	  int label=Integer.parseInt(ss[1]);
      	  double press=Double.parseDouble(ss[3]);
      	  model.getBoundaryCondition().addPressure
      	    (new Pressure(label,ss[2],press));
      	}
      	else if(keyWord.equals("HTC")){
      	  int label=Integer.parseInt(ss[1]);
      	  double htc=Double.parseDouble(ss[3]);
      	  double temp=Double.parseDouble(ss[4]);
      	  model.getBoundaryCondition().addHTBound
      	    (new HeatTransferBound(label,ss[2],htc,temp));
      	}
      	else if(keyWord.equals("Temperature")){
      	  int label=Integer.parseInt(ss[1]);
      	  double temp=Double.parseDouble(ss[2]);
      	  model.getBoundaryCondition().addTemperature
      	    (new Temperature(label,temp));
      	}
      }
    }
    model.init();
  }
}
